from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import os
import json
from uuid import uuid4
from copy import deepcopy

import compas

from compas.data.encoders import DataEncoder
from compas.data.encoders import DataDecoder


# ==============================================================================
# If you ever feel tempted to use ABCMeta in your code: don't, just DON'T.
# Assigning __metaclass__ = ABCMeta to a class causes a severe memory leak/performance
# degradation on IronPython 2.7.

# See these issues for more details:
# - https://github.com/compas-dev/compas/issues/562
# - https://github.com/compas-dev/compas/issues/649

# ==============================================================================


class Data(object):
    """Abstract base class for all COMPAS data objects.

    Parameters
    ----------
    name : str, optional
        The name of the object.

    Attributes
    ----------
    dtype : str, read-only
        The type of the object in the form of a fully qualified module name and a class name, separated by a forward slash ("/").
        For example: ``"compas.datastructures/Mesh"``.
    data : dict
        The representation of the object as a dictionary containing only built-in Python data types.
        The structure of the dict is described by the data schema.
    jsonstring : str, read-only
        The object's data dict in JSON string format.
    guid : str, read-only
        The globally unique identifier of the object.
        The guid is generated with ``uuid.uuid4()``.
    name : str
        The name of the object.
        This name is not necessarily unique and can be set by the user.
        The default value is the object's class name: ``self.__class__.__name__``.

    Notes
    -----
    Objects created from classes that implement this data class
    can be serialized to JSON and unserialized without loss of information using:

    * :func:`compas.data.json_dump`
    * :func:`compas.data.json_dumps`
    * :func:`compas.data.json_load`
    * :func:`compas.data.json_loads`

    To implement this data class,
    it is sufficient for the deriving class to define the "getter" and "setter"
    of the data property: :attr:`compas.data.Data.data`.

    Examples
    --------
    >>> from compas.data import Data
    >>> class Point(Data):
    ...     def __init__(self, x, y, z):
    ...         super().__init__()
    ...         self.x = x
    ...         self.y = y
    ...         self.z = z
    ...     @property
    ...     def data(self):
    ...         return {'x': self.x, 'y': self.y, 'z': self.z}
    ...     @data.setter
    ...     def data(self, data):
    ...         self.x = data['x']
    ...         self.y = data['y']
    ...         self.z = data['z']
    ...
    >>> a = Point(1.0, 0.0, 0.0)
    >>> a.guid                 # doctest: +SKIP
    UUID('1ddad2fe-6716-4e30-a5ae-8ed7cad892c4')
    >>> a.name
    'Point'
    >>> a.data
    {'x': 1.0, 'y': 0.0, 'z': 0.0}

    >>> from compas.data import json_dumps, json_loads
    >>> s = json_dumps(a)
    >>> b = json_loads(s)                                     # doctest: +SKIP
    >>> a is b                                                # doctest: +SKIP
    False
    >>> a == b                                                # doctest: +SKIP
    True

    """

    def __init__(self, name=None):
        self._guid = None
        self._name = None
        self._jsondefinitions = None
        self._JSONSCHEMA = None
        self._jsonvalidator = None
        if name:
            self.name = name

    def __getstate__(self):
        """Return the object data for state serialization with older pickle protocols."""
        return {'__dict__': self.__dict__, 'dtype': self.dtype, 'data': self.data}

    def __setstate__(self, state):
        """Assign a deserialized state to the object data to support older pickle protocols."""
        self.__dict__.update(state['__dict__'])
        self.data = state['data']

    @property
    def DATASCHEMA(self):
        """schema.Schema : The schema of the data of this object."""
        raise NotImplementedError

    @property
    def JSONSCHEMANAME(self):
        """str : The schema of the data of this object in JSON format."""
        raise NotImplementedError

    @property
    def JSONSCHEMA(self):
        """dict : The schema of the JSON representation of the data of this object."""
        if not self._JSONSCHEMA:
            schema_filename = '{}.json'.format(self.JSONSCHEMANAME.lower())
            schema_path = os.path.join(os.path.dirname(__file__), 'schemas', schema_filename)
            with open(schema_path, 'r') as fp:
                self._JSONSCHEMA = json.load(fp)
        return self._JSONSCHEMA

    @property
    def jsondefinitions(self):
        """dict : Reusable schema definitions."""
        if not self._jsondefinitions:
            schema_path = os.path.join(os.path.dirname(__file__), 'schemas', 'compas.json')
            with open(schema_path, 'r') as fp:
                self._jsondefinitions = json.load(fp)
        return self._jsondefinitions

    @property
    def jsonvalidator(self):
        """jsonschema.Draft7Validator : JSON schema validator for draft 7."""
        if not self._jsonvalidator:
            from jsonschema import RefResolver, Draft7Validator
            resolver = RefResolver.from_schema(self.jsondefinitions)
            self._jsonvalidator = Draft7Validator(self.JSONSCHEMA, resolver=resolver)
        return self._jsonvalidator

    @property
    def dtype(self):
        return '{}/{}'.format('.'.join(self.__class__.__module__.split('.')[:2]), self.__class__.__name__)

    @property
    def data(self):
        raise NotImplementedError

    @data.setter
    def data(self, data):
        raise NotImplementedError

    @property
    def jsonstring(self):
        return compas.json_dumps(self.data)

    @property
    def guid(self):
        if not self._guid:
            self._guid = uuid4()
        return self._guid

    @property
    def name(self):
        if not self._name:
            self._name = self.__class__.__name__
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @classmethod
    def from_data(cls, data):
        """Construct an object of this type from the provided data.

        Parameters
        ----------
        data : dict
            The data dictionary.

        Returns
        -------
        :class:`compas.data.Data`
            An instance of this object type if the data contained in the dict has the correct schema.

        """
        obj = cls()
        obj.data = data
        return obj

    def to_data(self):
        """Convert an object to its native data representation.

        Returns
        -------
        dict
            The data representation of the object as described by the schema.

        """
        return self.data

    @classmethod
    def from_json(cls, filepath):
        """Construct an object from serialized data contained in a JSON file.

        Parameters
        ----------
        filepath : path string | file-like object | URL string
            The path, file or URL to the file for serialization.

        Returns
        -------
        :class:`compas.data.Data`
            An instance of this object type if the data contained in the JSON file has the correct schema.

        """
        data = compas.json_load(filepath)
        return cls.from_data(data)

    def to_json(self, filepath, pretty=False):
        """Serialize the data representation of an object to a JSON file.

        Parameters
        ----------
        filepath : path string or file-like object
            The path or file-like object to the file containing the data.
        pretty : bool, optional
            If True, serialize to a "pretty", human-readable representation.

        Returns
        -------
        None

        """
        compas.json_dump(self.data, filepath, pretty)

    @classmethod
    def from_jsonstring(cls, string):
        """Construct an object from serialized data contained in a JSON string.

        Parameters
        ----------
        string : str
            The object as a JSON string.

        Returns
        -------
        :class:`compas.data.Data`
            An instance of this object type if the data contained in the JSON file has the correct schema.

        """
        data = compas.json_loads(string)
        return cls.from_data(data)

    def to_jsonstring(self, pretty=False):
        """Serialize the data representation of an object to a JSON string.

        Parameters
        ----------
        pretty : bool, optional
            If True serialize a pretty representation of the data.

        Returns
        -------
        str
            The object's data dict in JSON string format.

        """
        return compas.json_dumps(self.data, pretty)

    def copy(self, cls=None):
        """Make an independent copy of the data object.

        Parameters
        ----------
        cls : Type[:class:`compas.data.Data`], optional
            The type of data object to return.
            Defaults to the type of the current data object.

        Returns
        -------
        :class:`compas.data.Data`
            An independent copy of this object.

        """
        if not cls:
            cls = type(self)
        return cls.from_data(deepcopy(self.data))

    def validate_data(self):
        """Validate the object's data against its data schema.

        Returns
        -------
        dict
            The validated data.

        Raises
        ------
        schema.SchemaError

        """
        import schema
        try:
            data = self.DATASCHEMA.validate(self.data)
        except schema.SchemaError as e:
            print("Validation against the data schema of this object failed.")
            raise e
        return data

    def validate_json(self):
        """Validate the object's data against its json schema.

        Returns
        -------
        str
            The validated JSON representation of the data.

        Raises
        ------
        jsonschema.exceptions.ValidationError

        """
        import jsonschema
        jsonstring = json.dumps(self.data, cls=DataEncoder)
        jsondata = json.loads(jsonstring, cls=DataDecoder)
        try:
            self.jsonvalidator.validate(jsondata)
        except jsonschema.exceptions.ValidationError as e:
            print("Validation against the JSON schema of this object failed.")
            raise e
        return jsonstring



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.robots.resources.basic import AbstractMeshLoader
from compas.robots.resources.basic import _mesh_import

__all__ = ['GithubPackageMeshLoader']

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen


class GithubPackageMeshLoader(AbstractMeshLoader):
    """Loads resources stored in Github.

    Attributes
    ----------
    repository : str
        Repository name including organization,
        e.g. ``ros-industrial/abb``.
    support_package : str
        Name of the support package containing URDF, Meshes
        and additional assets, e.g. 'abb_irb4400_support'
    branch : str
        Branch name, defaults to ``main``.
    relative_path : str
        Relative path of the support package within the repository.
        If the repository itself is the support package, set
        ``relative_path`` to ``'.'``.  Defaults to ``support_package``
    """

    HOST = 'https://raw.githubusercontent.com'

    def __init__(self, repository, support_package, branch='main', relative_path=None):
        super(GithubPackageMeshLoader, self).__init__()
        self.repository = repository
        self.support_package = support_package
        self.branch = branch
        self.schema_prefix = 'package://' + self.support_package + '/'
        self.relative_path = support_package if relative_path is None else relative_path

    def build_url(self, file):
        """Returns the corresponding url of the file.

        Parameters
        ----------
        file : str
            File name. Following convention, the file should reside
            inside a ``urdf`` folder.

        Returns
        -------
        str
            The file's url.
        """
        relative_path_component = None if self.relative_path == '.' else self.relative_path
        url_components = [
            GithubPackageMeshLoader.HOST,
            self.repository,
            self.branch,
            relative_path_component,
            file
        ]
        return '/'.join(filter(None, url_components))

    def load_urdf(self, file):
        """Load a URDF file from a Github support package repository.

        Parameters
        ----------
        file : str
            File name. Following convention, the file should reside
            inside a ``urdf`` folder.
        """
        url = self.build_url('urdf/{}'.format(file))
        return urlopen(url)

    def can_load_mesh(self, url):
        """Determine whether this loader can load a given mesh URL.

        Parameters
        ----------
        url : str
            Mesh URL.

        Returns
        -------
        bool
            ``True`` if the URL uses the ``package://` scheme and the package name
            matches the specified in the constructor, otherwise ``False``.
        """
        return url.startswith(self.schema_prefix)

    def load_mesh(self, url):
        """Loads a mesh from a Github repository URL.

        .. deprecated:: 1.13.3
            Use :meth:`load_meshes` instead.

        Parameters
        ----------
        url : str
            Mesh location

        Returns
        -------
        :class:`compas.datastructures.Mesh`
            Instance of a mesh.
        """
        return self.load_meshes(url)[0]

    def load_meshes(self, url):
        """Load meshes from the given URL.

        A single mesh file can contain multiple meshes depending on the format.

        Parameters
        ----------
        url : str
            Mesh URL

        Returns
        -------
        list[:class:`compas.datastructures.Mesh`]
            List of meshes.
        """
        _prefix, path = url.split(self.schema_prefix)
        url = self.build_url(path)

        return _mesh_import(url, url)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino
from compas.artists import PrimitiveArtist
from .artist import RhinoArtist


class LineArtist(RhinoArtist, PrimitiveArtist):
    """Artist for drawing lines.

    Parameters
    ----------
    line : :class:`compas.geometry.Line`
        A COMPAS line.
    layer : str, optional
        The layer that should contain the drawing.
    **kwargs : dict, optional
        Additional keyword arguments.
        For more info, see :class:`RhinoArtist` and :class:`PrimitiveArtist`.

    """

    def __init__(self, line, layer=None, **kwargs):
        super(LineArtist, self).__init__(primitive=line, layer=layer, **kwargs)

    def draw(self, show_points=False):
        """Draw the line.

        Parameters
        ----------
        show_points : bool, optional
            If True, draw the start and end point of the line.

        Returns
        -------
        list[System.Guid]
            The GUIDs of the created Rhino objects.

        """
        start = list(self.primitive.start)
        end = list(self.primitive.end)
        guids = []
        if show_points:
            points = [
                {'pos': start, 'color': self.color, 'name': self.primitive.name},
                {'pos': end, 'color': self.color, 'name': self.primitive.name}
            ]
            guids += compas_rhino.draw_points(points, layer=self.layer, clear=False, redraw=False)
        lines = [{'start': start, 'end': end, 'color': self.color, 'name': self.primitive.name}]
        guids += compas_rhino.draw_lines(lines, layer=self.layer, clear=False, redraw=False)
        return guids



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import re
import random
import json
import copy


__all__ = ['ga']


TPL = """
GA summary
==========
- fitness function name: {}
- fitnes function type : {}
- number of generations : {}
- number of individuals : {}
- number of variables : {}
- optimal individual : {}
- optimal fitness value : {}
"""


def ga(fit_function,
       fit_type,
       num_var,
       boundaries,
       num_gen=100,
       num_pop=100,
       num_elite=10,
       mutation_probability=0.01,
       n_cross=1,
       num_bin_dig=None,
       num_pop_init=None,
       num_gen_init_pop=None,
       start_from_gen=False,
       min_fit=None,
       fit_name=None,
       fargs=None,
       fkwargs=None,
       output_path=None,
       input_path=None,
       print_refresh=1):
    """Genetic Algorithm optimisation.

    Parameters
    ----------
    fit_function : callable
        The function used by the :class'GA' to determine the fitness value. The function
        must have as a first argument a list of variables that determine the
        fitness value. Other arguments and keyword arguments can be used to feed
        the function relevant data.
    fit_type : str
        String that indicates if the fitness function is to be minimized or maximized.
        "min" for minimization and "max" for maximization.
    num_var :  int
        The number of variables used by the fitness function.
    boundaries : list
        The minimum and vaximum values each variable is allowed to have. Must be
        a ``num_var`` long list of tuples in the form [(min, max),...].
    num_gen : int, optional [100]
        The maximum number of generations.
    num_pop : int, optional [100]
        The number of individuals in the population. Must be an even number.
    num_elite : int, optional [10]
        The number of individuals in the elite population. Must be an even number.
    mutation_probablity : float, optional [0.001]
        Float from 0 to 1. Percentage of genes that will be mutated.
    n_cross: int, optional [1]
        Number of crossover points used in the crossover operator.
    num_bin_dig : list, optional [None]
        Number of genes used to codify each variable. Must be a ``num_var`` long
        list of intergers. If None is given, each variable will be coded with a
        8 digit binary number, corresponding to 256 steps.
    num_pop_init : int, optional [None]
        The number of individuals in the population for the first ``num_gen_init_pop``
        generations.
    num_gen_init_pop : int, optional
        The number of generations to keep a ``num_pop_init`` size population for.
    start_from_get : int, optional [None]
        The generation number to restart a previous optimization process.
    min_fit : float, optional [None]
        A target fitness value. If the GA finds a solution with a fitness value
        equal or better than ``min_fit``, the optimization is stopped.
    fit_name : str, optional [None]
        The name of the optimisation. If None is given, the name of the fitness
        function is used.
    fargs : list, optional [None]
        Arguments fo be fed to the fitness function.
    fkwargs : dict, optional [None]
        Keyword arguments to be fed to the fitness function.
    output_path : str, optional [None]
        Path for the optimization result files.
    input_path : str, optional [None]
        Path to the fitness function file.
    print_refresh : int
        Print current generation summary every ``print_refresh`` generations.

    Returns
    -------
    ga_ : object
        The resulting :class'GA' instance.

    Notes
    -----
    For more info, see [1]_.

    References
    ----------
    .. [1] Holland, J. H., *Adaptation in Natural and Artificial Systems*, 1st edn,
           The University of Michigan, Ann Arbor, 1975.

    Examples
    --------
    >>>
    """
    ga_ = GA()

    ga_.fit_name = fit_name or fit_function.__name__
    ga_.fit_type = fit_type
    ga_.num_gen = num_gen
    ga_.num_pop = num_pop
    ga_.num_pop_init = num_pop_init
    ga_.num_gen_init_pop = num_gen_init_pop
    ga_.num_elite = num_elite
    ga_.num_var = num_var
    ga_.mutation_probability = mutation_probability
    ga_.ncross = n_cross
    ga_.start_from_gen = start_from_gen
    ga_.min_fit = min_fit
    ga_.boundaries = boundaries
    ga_.num_bin_dig = num_bin_dig or [8] * num_var
    ga_.max_bin_dig = max(ga_.num_bin_dig)
    ga_.total_bin_dig = sum(ga_.num_bin_dig)
    ga_.fargs = fargs or {}
    ga_.fkwargs = fkwargs or {}
    ga_.fit_function = fit_function
    ga_.output_path = output_path or ''
    ga_.input_path = input_path or ''
    ga_.print_refresh = print_refresh
    ga_.ga_optimize()
    return ga_


class GA(object):
    """This class contains a binary coded, single objective genetic algorithm.

    Attributes
    ----------
    best_fit : float
        The fitness value of the best performing solution for the current generation.
    best_individual_index: int
        The index of the best performing individual for the current generation.
    boundaries : dict
        This dictionary contains all the max and min bounds for each optimization variable.
        ``GA.boundaries[index] = [min,max]``.
    current_pop : dict
        This dictionary contains the coded, decoded and scaled population of the current
        generation, as well as their fitness values.
    elite_pop : dict
        This dictionary contains the coded, decoded and scaled data for the elite
        population of the current generation, as well as their fitness values.
    end_gen : int
        The maximum number of generations the GA is allowed to run.
    fit_function: function
        The fitness function.
    fit_name : str
        The name of the python file containing the fitness function (without extension).
    fit_type : str
        String that indicates if the fitness function is to be minimized or maximized.
        "min" for minimization and "max" for maximization.
    input_path: str
        Path to the fitness function file.
    fkwargs : dict
        This dictionary will be passed as a keyword argument to all fitness functions.
        It can be used to pass required data, objects, that are not related to the
        optimmization variables but are required to run the fitness function.
    max_bin_digit : int
        The maximum number of binary digits that are used to code a variable values.
        The number of binary digits assigned to code a variable determine the number
        of discrete steps inside the variable bounds. For example, an 8 digit binary
        number will produce 256 steps.
    min_fit : float
        An end condition related to fitness value. If it is set, the GA will stop
        when any individual achieves a fitness equal or better that ``GA.min_fit``. If
        it is not set, the GA will end after ``GA.num_gen`` generations.
    min_fit_flag : bool
        Flag the GA uses to determine if the ``GA.min_fit`` value has been achieved.
    mutation_probability : float
        Determines the probability that the mutation operator will mutate each gene.
        For each gene a random number ``x`` between 0 and 1 is generated, if ``x``
        is higher than ``GA.mutation_probability`` it will be mutated.
    num_bin_dig : list
        List of the number of binary digits for each variable. The number of binary
        digits assigned to code a variable determine the number of discrete steps
        inside the variable bounds. For example, an 8 digit binary number will
        produce 256 steps.
    num_elite : int
        The number of top performing individuals in the population that are not subject
        to genetic operators, but are simply copied to the next generation.
    num_gen : int
        The number of genertions the GA will run.
    num_pop : int
        The number of individuals per generation.
    num_var : int
        The number of variables in the optimization problem.
    output_path : str
        The path to which the GA outputs result files.
    start_from_gen : int
        The generation from which the GA will restart. If this number is given, the GA
        will look for generation output files in the ``GA.input_path`` and if found,
        the GA will resume optimization from the ``GA.start_from_gen`` generation.
    total_bin_dig : int
        The total number of binary digits. It is the sum of the ``GA.num_bin_dig`` of
        all variables.
    ind_fit_dict : dict
        This dictionary keeps track of already evaluated solutions to avoid dupplicate
        fitness function calls.

    """

    def __init__(self):
        """ Initializes the GA object."""

        self.fkwargs = {}
        self.fargs = {}
        self.best_fit = None
        self.best_individual_index = None
        self.boundaries = {}
        self.current_pop = {'binary': [], 'decoded': [], 'scaled': [], 'fit_value': []}
        self.elite_pop = {'binary': [], 'decoded': [], 'scaled': [], 'fit_value': []}
        self.end_gen = None
        self.fit_function = None
        self.fit_name = ''
        self.fit_type = None
        self.input_path = None
        self.max_bin_dig = []
        self.min_fit = None
        self.min_fit_flag = False
        self.mutation_probability = 0
        self.n_cross = 1
        self.num_bin_dig = 0
        self.num_elite = 0
        self.num_gen = 0
        self.num_gen_init_pop = 1
        self.num_pop = 0
        self.num_pop_init = None
        self.num_pop_temp = None
        self.num_var = 0
        self.output_path = []
        self.start_from_gen = False
        self.total_bin_dig = 0
        self.check_diversity = False
        self.ind_fit_dict = {}
        self.print_refresh = 1

    def __str__(self):
        """Compile a summary of the GA."""
        fit_name = self.fit_name
        fit_type = self.fit_type
        num_gen = self.num_gen
        num_pop = self.num_pop
        num_var = self.num_var
        best = self.best_individual_index, self.current_pop['scaled'][self.best_individual_index]
        try:
            fit = self.current_pop['fit_value'][self.best_individual_index]
        except(Exception):
            fit = None
        return TPL.format(fit_name, fit_type, num_gen, num_pop, num_var, best, fit)

    def summary(self):
        """Print a summary of the GA."""
        print(self)

    def ga_optimize(self):
        """ This is the main optimization function, this function permorms the GA optimization,
        performing all genetic operators.
        """
        self.write_ga_json_file()

        if self.num_pop_init:
            self.num_pop_temp = copy.deepcopy(self.num_pop)
            self.num_pop = self.num_pop_init

        if self.start_from_gen:
            self.current_pop = self.get_pop_from_pop_file(self.start_from_gen)
            start_gen_number = self.start_from_gen + 1
        else:
            self.current_pop['binary'] = self.generate_random_bin_pop()
            start_gen_number = 0

        for generation in range(start_gen_number, self.num_gen):

            self.current_pop['decoded'] = self.decode_binary_pop(self.current_pop['binary'])
            self.current_pop['scaled'] = self.scale_population(self.current_pop['decoded'])

            if generation == 0:
                num = self.num_pop
                self.current_pop['fit_value'] = [[]] * num
            else:
                num = self.num_pop - self.num_elite

            for i in range(num):
                self.current_pop['fit_value'][i] = self.evaluate_fitness(i)

            if self.num_pop_init and generation >= self.num_gen_init_pop:
                self.num_pop = self.num_pop_temp
                self.current_pop = self.select_elite_pop(self.current_pop, num_elite=self.num_pop)
            self.write_out_file(generation)

            if self.min_fit:
                self.update_min_fit_flag()
            else:
                self.get_best_fit()
            if generation % self.print_refresh == 0:
                print('generation ', generation, ' best fit ', self.best_fit, 'min fit', self.min_fit)

            if self.check_diversity:
                print('num repeated individuals', self.check_pop_diversity())
            if generation < self.num_gen - 1 and self.min_fit_flag is False:
                self.elite_pop = self.select_elite_pop(self.current_pop)
                self.tournament_selection()  # n-e
                self.create_mating_pool()  # n-e
                self.npoint_crossover()  # n-e
                self.random_mutation()  # n-e
                self.add_elite_to_current()  # n

            else:
                self.end_gen = generation
                self.get_best_individual_index()
                self.write_ga_json_file()
                print(self)
                break

    def evaluate_fitness(self, index):
        chromo = ''.join(str(y) for x in self.current_pop['binary'][index] for y in x)
        fit = self.ind_fit_dict.setdefault(chromo, None)
        if not fit:
            fit = self.fit_function(self.current_pop['scaled'][index], *self.fargs, **self.fkwargs)
            self.ind_fit_dict[chromo] = fit
        return fit

    def check_pop_diversity(self):
        seen = []
        all_ = []
        for ind in self.current_pop['scaled']:
            if ind not in seen:
                seen.append(ind)
            all_.append(ind)
        return len(all_) - len(seen)

    def decode_binary_pop(self, bin_pop):
        """Decodes the binary population, from binary to unscaled variable values

        Parameters
        ----------
        bin_pop: list
            The binary population list.

        Returns
        -------
        decoded_pop:
            The decoded population list.
        """
        decoded_pop = [[[]] * self.num_var for i in range(self.num_pop)]
        for j in range(self.num_pop):
            for i in range(self.num_var):
                value = 0
                chrom = bin_pop[j][i]
                for u, gene in enumerate(chrom):
                    if gene == 1:
                        value = value + 2**u
                decoded_pop[j][i] = value
        return decoded_pop

    def generate_random_bin_pop(self):
        """ Generates random binary population of ``GA.num_pop`` size.

        Returns
        -------
        random_bin_pop: list
            A list containing a random binary population.
        """
        random_bin_pop = [[[]] * self.num_var for i in range(self.num_pop)]
        for j in range(self.num_pop):
            for i in range(self.num_var):
                random_bin_pop[j][i] = [random.randint(0, 1) for u in range(self.num_bin_dig[i])]
        return random_bin_pop

    def scale_population(self, decoded_pop):
        """Scales the decoded population, variable values are scaled according to each
        of their bounds contained in ``GA.boundaries``.

        Parameters
        ----------
        decoded_pop: list
            The decoded population list.

        Returns
        -------
        scaled_pop: list
            The scaled ppopulation list.
        """
        scaled_pop = [[[]] * self.num_var for i in range(self.num_pop)]
        for j in range(self.num_pop):
            for i in range(self.num_var):
                maxbin = float((2 ** self.num_bin_dig[i]) - 1)
                scaled_pop[j][i] = self.boundaries[i][0] + (self.boundaries[i][1] - self.boundaries[i][0]) * decoded_pop[j][i] / maxbin
        return scaled_pop

    def tournament_selection(self):
        """Performs the tournament selection operator on the current population.
        """
        pop_a = []
        pop_b = []
        indices = range(self.num_pop)
        pop_a = random.sample(indices, self.num_pop - self.num_elite)
        pop_b = random.sample(indices, self.num_pop - self.num_elite)
        self.mp_indices = []
        for i in range(self.num_pop - self.num_elite):
            fit_a = self.current_pop['fit_value'][pop_a[i]]
            fit_b = self.current_pop['fit_value'][pop_b[i]]
            if self.fit_type == 'min':
                if fit_a < fit_b:
                    self.mp_indices.append(pop_a[i])
                else:
                    self.mp_indices.append(pop_b[i])
            elif self.fit_type == 'max':
                if fit_a > fit_b:
                    self.mp_indices.append(pop_a[i])
                else:
                    self.mp_indices.append(pop_b[i])

    def select_elite_pop(self, pop, num_elite=None):
        """Saves the elite population in the elite population dictionary

        Parameters
        ----------
        pop: dict
            A population dictionary

        Returns
        -------
        elite_pop: dict
            The elite population dictionary.
        """
        if self.fit_type == 'min':
            sorted_ind = self.get_sorting_indices(self.current_pop['fit_value'])
        elif self.fit_type == 'max':
            sorted_ind = self.get_sorting_indices(self.current_pop['fit_value'], reverse=True)
        else:
            raise ValueError('User selected fit_type is wrong. Use "min" or "max" only')
        if not num_elite:
            num_elite = self.num_elite

        elite_pop = {'binary': [], 'decoded': [], 'scaled': [], 'fit_value': []}
        for i in range(num_elite):
            elite_pop['binary'].append(pop['binary'][sorted_ind[i]])
            elite_pop['decoded'].append(pop['decoded'][sorted_ind[i]])
            elite_pop['scaled'].append(pop['scaled'][sorted_ind[i]])
            elite_pop['fit_value'].append(pop['fit_value'][sorted_ind[i]])
        return elite_pop

    def get_sorting_indices(self, floats, reverse=False):
        """Reurns the indices that would sort a list of floats. If floats are
        repeated in the list, only one instance is considered. The index of
        repeaded floats are included in the end of the index list.

        Parameters
        ----------
        floats: list
            The list of floats to be sorted.
        reverse: bool
            If true the sorting will be done from top to bottom.

        Returns
        -------
        sorting_index: list
            The list of indices that would sort the given list of floats.
        """
        l_ = []
        if reverse:
            x = float('-inf')
        else:
            x = float('inf')
        for i in floats:
            if i in l_:
                l_.append(x)
            else:
                l_.append(i)
        sorting_index = [i for (v, i) in sorted((v, i) for (i, v) in enumerate(l_))]
        if reverse is True:
            sorting_index = list(reversed(sorting_index))
        return sorting_index

    def create_mating_pool(self):
        """Creates two lists of cromosomes to be used by the crossover operator.
        """
        self.mating_pool_a = []
        self.mating_pool_b = []
        for i in range(int((self.num_pop - self.num_elite) / 2)):
            chrom_a = []
            chrom_b = []
            for j in range(self.num_var):
                chrom_a += self.current_pop['binary'][self.mp_indices[i]][j]
                chrom_b += self.current_pop['binary'][self.mp_indices[i + (int((self.num_pop - self.num_elite) / 2))]][j]
            self.mating_pool_a.append(chrom_a)
            self.mating_pool_b.append(chrom_b)

    def simple_crossover(self):
        """Performs the simple crossover operator. Individuals in ``GA.mating_pool_a`` are
        combined with individuals in ``GA.mating_pool_b`` using a single, randomly selected
        crossover point.
        """
        self.current_pop = {'binary': [], 'decoded': [], 'scaled': [], 'fit_value': []}
        self.current_pop['binary'] = [[[]] * self.num_var for i in range(self.num_pop)]
        for j in range(int((self.num_pop - self.num_elite) / 2)):
            cross = random.randint(1, self.total_bin_dig - 1)
            a = self.mating_pool_a[j]
            b = self.mating_pool_b[j]
            c = a[:cross] + b[cross:]
            d = b[:cross] + a[cross:]

            for i in range(self.num_var):
                variable_a = c[:self.num_bin_dig[i]]
                variable_b = d[:self.num_bin_dig[i]]
                del c[:self.num_bin_dig[i]]
                del d[:self.num_bin_dig[i]]
                self.current_pop['binary'][j][i] = variable_a
                self.current_pop['binary'][j + (int((self.num_pop - self.num_elite) / 2))][i] = variable_b

    def npoint_crossover(self):
        """Performs the n-point crossover operator. Individuals in ``GA.mating_pool_a`` are
        combined with individuals in ``GA.mating_pool_b`` using ne, randomly selected
        crossover points.
        """
        self.current_pop = {'binary': [], 'decoded': [], 'scaled': [], 'fit_value': []}
        self.current_pop['binary'] = [[[]] * self.num_var for i in range(self.num_pop)]
        for j in range(int((self.num_pop - self.num_elite) / 2)):
            a = self.mating_pool_a[j]
            b = self.mating_pool_b[j]
            cross_list = sorted(random.sample(range(1, self.total_bin_dig - 1), self.n_cross))
            for cross in cross_list:
                c = a[:cross] + b[cross:]
                d = b[:cross] + a[cross:]
                a = d
                b = c
            for i in range(self.num_var):
                variable_a = a[:self.num_bin_dig[i]]
                variable_b = b[:self.num_bin_dig[i]]
                self.current_pop['binary'][j][i] = variable_a
                self.current_pop['binary'][j + (int((self.num_pop - self.num_elite) / 2))][i] = variable_b

    def random_mutation(self):
        """This mutation operator replaces a gene from 0 to 1 or viceversa
        with a probability of ``GA.mutation_probability``.
        """
        for i in range(self.num_pop - self.num_elite):
            for j in range(self.num_var):
                for u in range(self.num_bin_dig[j]):
                    random_value = random.random()
                    if random_value < (self.mutation_probability):
                        if self.current_pop['binary'][i][j][u] == 0:
                            self.current_pop['binary'][i][j][u] = 1
                        else:
                            self.current_pop['binary'][i][j][u] = 0

    def code_decoded(self, decoded_pop):
        """Returns a binary coded population from a decoded population

        Parameters
        ----------
        decoded_pop: dict
            The decoded population dictionary to be coded

        Returns
        -------
        binary_pop: dict
            The binary population dictionary.
        """
        binary_pop = [[[]] * self.num_var for i in range(self.num_pop)]
        for i in range(self.num_pop):
            for j in range(self.num_var):
                bin_list = []
                temp_bin = bin(decoded_pop[i][j])[2:]
                temp_bin = temp_bin[::-1]
                digit_dif = self.num_bin_dig[j] - len(temp_bin)
                for h in range(digit_dif):
                    temp_bin = temp_bin + '0'
                for k in range(self.num_bin_dig[j]):
                    bin_list.append(int(temp_bin[k]))
                binary_pop[i][j] = bin_list
        return binary_pop

    def unscale_pop(self, scaled_pop):
        """Returns an unscaled population from a scaled one. The variable values are scaled
        from 0 to x, where x is the highest number described by the number of binary digits
        used to encode that variable. For example, if ``GA.num_bin_dig`` for a variable is 8, that
        variable will be scaled back from its bounds to its corresponding value from 0 to 255.

        Parameters
        ----------
        scaled_pop: dict
            the scaled population dictionary.

        Returns
        -------
        unscaled_pop: dict
            The unscaled population dictionary.
        """
        unscaled_pop = {}

        for i in range(self.num_pop):
            unscaled_pop[i] = {}
            for j in range(self.num_var):
                bin_dig = self.num_bin_dig[j]
                bounds = self.boundaries[j]
                max_unscaled_value = self.get_max_value_from_bin_big(bin_dig)
                dom = abs(bounds[1] - bounds[0])
                value_s = scaled_pop[i][j]
                value = (value_s - bounds[0]) / float(dom)
                unscaled = int(value * max_unscaled_value)
                unscaled_pop[i][j] = unscaled

        return unscaled_pop

    def get_max_value_from_bin_big(self, bin_dig):
        """Returns the highest number described by a ``GA.bin_dig`` long binary number.

        Parameters
        ----------
        bin_dig: int
            The number of digits in the binary number.

        Returns
        -------
        value: int
            The highest number described by a ``GA.bin_dig`` long binary number.
        """
        binary = ''
        for i in range(bin_dig):
            binary += '1'
        value = 0
        for i in range(bin_dig):
            value = value + 2**i
        return value

    def write_out_file(self, generation):
        """Writes the population data for a given generation.

        Parameters
        ----------
        generation: int
            The generation number.
        """
        filename = 'generation_' + "%05d" % generation + '_population' + ".txt"
        pf_file = open(self.output_path + (str(filename)), "w")
        pf_file.write('Generation \n')
        pf_file.write(str(generation) + '\n')
        pf_file.write('\n')

        pf_file.write('Number of individuals per generation\n')
        pf_file.write(str(self.num_pop))
        pf_file.write('\n')
        pf_file.write('\n')

        pf_file.write('Population scaled variables \n')
        for i in range(self.num_pop):
            pf_file.write(str(i) + ',')
            for f in range(self.num_var):
                pf_file.write(str(self.current_pop['scaled'][i][f]))
                pf_file.write(',')
            pf_file.write('\n')
        pf_file.write('\n')

        pf_file.write('Population fitness value \n')
        for i in range(self.num_pop):
            pf_file.write(str(i) + ',')
            pf_file.write(str(self.current_pop['fit_value'][i]))
            pf_file.write('\n')
        pf_file.write('\n')
        pf_file.write('\n')
        pf_file.close()

    def add_elite_to_current(self):
        """Adds the elite population to the current population dictionary.
        """
        self.current_pop['decoded'] = [[[]] * self.num_var for i in range(self.num_pop)]
        self.current_pop['decoded'] = [[[]] * self.num_var for i in range(self.num_pop)]
        self.current_pop['scaled'] = [[[]] * self.num_var for i in range(self.num_pop)]
        self.current_pop['fit_value'] = [[]] * self.num_pop
        for i in range(self.num_elite):
            self.current_pop['binary'][self.num_pop - self.num_elite + i] = self.elite_pop['binary'][i]
            self.current_pop['decoded'][self.num_pop - self.num_elite + i] = self.elite_pop['decoded'][i]
            self.current_pop['scaled'][self.num_pop - self.num_elite + i] = self.elite_pop['scaled'][i]
            self.current_pop['fit_value'][self.num_pop - self.num_elite + i] = self.elite_pop['fit_value'][i]

    def make_ga_input_data(self):
        """Returns a dictionary containing the most relavant genetic data present in the instance
        of ``GA``. This is the data required to restart a genetic optimization process or to
        launch a visualization using ``compas_ga.visualization.ga_visualization``.

        Returns
        -------
        data: dict
            A dictionary containing genetic data.
        """
        data = {'num_var': self.num_var,
                'num_pop': self.num_pop,
                'num_gen': self.num_gen,
                'boundaries': self.boundaries,
                'num_bin_dig': self.num_bin_dig,
                'mutation_probability': self.mutation_probability,
                'fit_name': self.fit_name,
                'fit_type': self.fit_type,
                'start_from_gen': self.start_from_gen,
                'max_bin_dig': self.max_bin_dig,
                'total_bin_dig': self.total_bin_dig,
                'output_path': self.output_path,
                'num_elite': self.num_elite,
                'min_fit': self.min_fit,
                'end_gen': self.end_gen,
                'best_individual_index': self.best_individual_index
                }
        return data

    def write_ga_json_file(self):
        """Writes a JSON file containing the most relevant data for GA optimization and
        visualization using ``compas_ga.visualization.ga_visualization``.
        """
        data = self.make_ga_input_data()
        filename = self.fit_name + '.json'
        with open(self.output_path + filename, 'w') as fh:
            json.dump(data, fh)

    def update_min_fit_flag(self):
        """Checks if the minimum desired fitness value has been achieved during optimization
        and saves the result in ``GA.min_fit_flag``.
        """
        values = self.current_pop['fit_value']
        if self.fit_type == 'min':
            self.best_fit = min(values)
            if self.best_fit <= self.min_fit:
                self.min_fit_flag = True
        elif self.fit_type == 'max':
            self.best_fit = max(values)
            if self.best_fit >= self.min_fit:
                self.min_fit_flag = True

    def get_best_fit(self):
        """Saves the best fitness value in ``GA.best_fit``
        """
        if self.fit_type == 'min':
            self.best_fit = min(self.current_pop['fit_value'])
        elif self.fit_type == 'max':
            self.best_fit = max(self.current_pop['fit_value'])

    def get_pop_from_pop_file(self, gen):
        """Reads the population file corresponding to the ``gen`` generation and returns
        the saved population data. The population file must be in ``GA.input_path``.

        Parameters
        ----------
        gen: int
            The generation index.

        Returns
        -------
        file_pop: dict
            The population dictionary contained in the file.
        """
        filename = 'generation_' + "%05d" % gen + '_population' + ".txt"
        filename = self.input_path + filename
        pf_file = open(filename, 'r')
        lines = pf_file.readlines()
        pf_file.close()
        file_pop = {}
        file_pop['scaled'] = [[[]] * self.num_var for i in range(self.num_pop)]
        file_pop['fit_value'] = [[[]] * self.num_var for i in range(self.num_pop)]

        for i in range(self.num_pop):
            line_scaled = lines[i + 7]
            line_fit = lines[i + 9 + self.num_pop]
            string_scaled = re.split(',', line_scaled)
            string_fit = re.split(',', line_fit)
            string_fit = string_fit[1]
            del string_scaled[-1]
            del string_scaled[0]
            scaled = [float(j) for j in string_scaled]
            fit_value = float(string_fit)
            file_pop['fit_value'][i] = fit_value
            for j in range(len(scaled)):
                file_pop['scaled'][i][j] = scaled[j]

        file_pop['decoded'] = self.unscale_pop(file_pop['scaled'])
        file_pop['binary'] = self.code_decoded(file_pop['decoded'])
        return file_pop

    def get_best_individual_index(self):
        """Saves the index of the best performing individual of the current population
         in ``GA.best_individual_index``.
        """
        fit_values = [self.current_pop['fit_value'][i] for i in range(len(self.current_pop['fit_value']))]
        if self.fit_type == 'min':
            indices = self.get_sorting_indices(fit_values)
        elif self.fit_type == 'max':
            indices = self.get_sorting_indices(fit_values, reverse=True)
        self.best_individual_index = indices[0]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import compas_rhino
from compas.artists import PrimitiveArtist
from .artist import RhinoArtist


class FrameArtist(RhinoArtist, PrimitiveArtist):
    """Artist for drawing frames.

    Parameters
    ----------
    frame: :class:`compas.geometry.Frame`
        A COMPAS frame.
    scale: float, optional
        Scale factor that controls the length of the axes.
    layer : str, optional
        The layer that should contain the drawing.
    **kwargs : dict, optional
        Additional keyword arguments.
        For more info, see :class:`RhinoArtist` and :class:`PrimitiveArtist`.

    Attributes
    ----------
    scale : float
        Scale factor that controls the length of the axes.
        Default is ``1.0``.
    color_origin : tuple of 3 int between 0 and 255
        Default is ``(0, 0, 0)``.
    color_xaxis : tuple of 3 int between 0 and 255
        Default is ``(255, 0, 0)``.
    color_yaxis : tuple of 3 int between 0 and 255
        Default is ``(0, 255, 0)``.
    color_zaxis : tuple of 3 int between 0 and 255
        Default is ``(0, 0, 255)``.

    """

    def __init__(self, frame, layer=None, scale=1.0, **kwargs):
        super(FrameArtist, self).__init__(primitive=frame, layer=layer, **kwargs)
        self.scale = scale or 1.0
        self.color_origin = (0, 0, 0)
        self.color_xaxis = (255, 0, 0)
        self.color_yaxis = (0, 255, 0)
        self.color_zaxis = (0, 0, 255)

    def draw(self):
        """Draw the frame.

        Returns
        -------
        list[System.Guid]
            The GUIDs of the created Rhino objects.

        """
        points = []
        lines = []
        origin = list(self.primitive.point)
        X = list(self.primitive.point + self.primitive.xaxis.scaled(self.scale))
        Y = list(self.primitive.point + self.primitive.yaxis.scaled(self.scale))
        Z = list(self.primitive.point + self.primitive.zaxis.scaled(self.scale))
        points = [{'pos': origin, 'color': self.color_origin}]
        lines = [
            {'start': origin, 'end': X, 'color': self.color_xaxis, 'arrow': 'end'},
            {'start': origin, 'end': Y, 'color': self.color_yaxis, 'arrow': 'end'},
            {'start': origin, 'end': Z, 'color': self.color_zaxis, 'arrow': 'end'}]
        guids = compas_rhino.draw_points(points, layer=self.layer, clear=False, redraw=False)
        guids += compas_rhino.draw_lines(lines, layer=self.layer, clear=False, redraw=False)
        return guids



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from .gltf import GLTF
from .gltf_content import GLTFContent
from .gltf_exporter import GLTFExporter
from .gltf_mesh import GLTFMesh
from .gltf_parser import GLTFParser
from .gltf_reader import GLTFReader

__all__ = [
    'GLTF',
    'GLTFContent',
    'GLTFMesh',
    'GLTFReader',
    'GLTFParser',
    'GLTFExporter',
]



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.utilities import geometric_key

import Rhino
import scriptcontext as sc


__all__ = ['volmesh_from_polysurfaces']


def volmesh_from_polysurfaces(cls, guids, precision=None):
    """Construct a volumetric mesh from given polysurfaces.

    Parameters
    ----------
    cls : :class:`compas.datastructures.VolMesh`
        The class of volmesh.
    guids : sequence[str or System.Guid]
        The *globally unique identifiers* of the polysurfaces.
    precision: string
        Precision of the polysurface connectivity.

    Returns
    -------
    :class:`compas.datastructures.Volmesh`
        The volumetric mesh object.

    Notes
    -----
    Essentially, this function does the following:

    * find each of the polysurfaces and check if they have a boundary representation (b-rep)
    * convert to b-rep and extract the edge loops
    * make a face of each loop by referring to vertices using their geometric keys
    * add a cell per brep
    * and add the faces of a brep to the cell
    * create a volmesh from the found vertices and cells

    """
    gkey_xyz = {}
    cells = []
    for guid in guids:
        cell = []
        obj = sc.doc.Objects.Find(guid)
        if not obj.Geometry.HasBrepForm:
            continue
        brep = Rhino.Geometry.Brep.TryConvertBrep(obj.Geometry)
        for loop in brep.Loops:
            curve = loop.To3dCurve()
            segments = curve.Explode()
            face = []
            sp = segments[0].PointAtStart
            ep = segments[0].PointAtEnd
            sp_gkey = geometric_key(sp, precision)
            ep_gkey = geometric_key(ep, precision)
            gkey_xyz[sp_gkey] = sp
            gkey_xyz[ep_gkey] = ep
            face.append(sp_gkey)
            face.append(ep_gkey)
            for segment in segments[1:-1]:
                ep = segment.PointAtEnd
                ep_gkey = geometric_key(ep, precision)
                face.append(ep_gkey)
                gkey_xyz[ep_gkey] = ep
            cell.append(face)
        cells.append(cell)
    gkey_index = dict((gkey, index) for index, gkey in enumerate(gkey_xyz))
    vertices = [list(xyz) for gkey, xyz in gkey_xyz.items()]
    cells = [[[gkey_index[gkey] for gkey in face] for face in cell] for cell in cells]
    return cls.from_vertices_and_cells(vertices, cells)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from collections import deque
from compas.topology import breadth_first_traverse


def vertex_coloring(adjacency):
    """Color the vertices of a graph such that no two colors are adjacent.

    Parameters
    ----------
    adjacency : dict[hashable, dict[hashable, None]] or dict[hashable, sequence[hashable]]
        An adjacency dictionary representing the connectivity of the graph
        by mapping nodes identifiers to neighbour identifiers.
        Examples of valid adjacency dicts are

        * ``{0: [1, 2, 3, 4], 1: [0], 2: [0], 3: [0], 4: [0]}``
        * ``{0: {1: None, 2: None, 3: None, 4: None}, 1: {0: None}, 2: {0: None}, 3: {0: None}, 4: {0: None}}``

    Returns
    -------
    dict[hashable, int]
        A dict mapping each node of the graph to a color number.

    Notes
    -----
    This algorithms works on any data structure that can be interpreted as a graph, e.g.
    networks, meshes, volmeshes, etc..

    For more info, see [1]_.

    References
    ----------
    .. [1] Chu-Carroll, M. *Graph Coloring Algorithms*.
           Available at: http://scienceblogs.com/goodmath/2007/06/28/graph-coloring-algorithms-1/.

    Warnings
    --------
    This is a greedy algorithm, so it might be slow for large networks.

    Examples
    --------
    >>> import compas
    >>> from compas.datastructures import Network
    >>> network = Network.from_obj(compas.get('lines.obj'))
    >>> key_color = vertex_coloring(network.adjacency)
    >>> key = network.get_any_node()
    >>> color = key_color[key]
    >>> any(key_color[nbr] == color for nbr in network.neighbors(key))
    False

    """
    key_to_color = {}
    key_to_degree = {key: len(adjacency[key]) for key in adjacency}
    vertices = sorted(adjacency.keys(), key=lambda key: key_to_degree[key])
    uncolored = deque(vertices[::-1])
    current_color = 0
    while uncolored:
        a = uncolored.popleft()
        key_to_color[a] = current_color
        colored_with_current = [a]
        for b in uncolored:
            if not any(b in adjacency[key] for key in colored_with_current):
                key_to_color[b] = current_color
                colored_with_current.append(b)
        for key in colored_with_current[1:]:
            uncolored.remove(key)
        current_color += 1
    return key_to_color


def connected_components(adjacency):
    """Identify the connected components of a graph.

    Parameters
    ----------
    adjacency : dict[hashable, dict[hashable, None]] | dict[hashable, sequence[hashable]]
        An adjacency dictionary representing the connectivity of the graph
        by mapping nodes identifiers to neighbour identifiers.
        Examples of valid adjacency dicts are

        * ``{0: [1, 2, 3, 4], 1: [0], 2: [0], 3: [0], 4: [0]}``
        * ``{0: {1: None, 2: None, 3: None, 4: None}, 1: {0: None}, 2: {0: None}, 3: {0: None}, 4: {0: None}}``

    Returns
    -------
    list[list[hashable]]
        A list of connected components,
        with each component a list of connected nodes.

    Examples
    --------
    >>> adjacency = {0: [1, 2], 1: [0, 2], 2: [0, 1], 3: []}
    >>> connected_components(adjacency)
    [[0, 1, 2], [3]]

    """
    tovisit = set(adjacency)
    components = []
    while tovisit:
        root = tovisit.pop()
        visited = breadth_first_traverse(adjacency, root)
        tovisit -= visited
        components.append(list(visited))
    return components



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

import Rhino

from ._primitives import circle_to_compas
from ._primitives import circle_to_rhino

from ._geometry import RhinoGeometry


class RhinoCircle(RhinoGeometry):
    """Wrapper for Rhino circles."""

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, geometry):
        """Set the geometry of the wrapper.

        Parameters
        ----------
        geometry : :rhino:`Rhino_Geometry_Circle` or :class:`compas.geometry.Circle`
            The geometry object defining a circle.

        Raises
        ------
        :class:`ConversionError`
            If the geometry cannot be converted to a box.
        """
        if not isinstance(geometry, Rhino.Geometry.Circle):
            geometry = circle_to_rhino(geometry)
        self._geometry = geometry

    def to_compas(self):
        """Convert to a COMPAS geometry object.

        Returns
        -------
        :class:`compas.geometry.Circle`
            A COMPAS circle.
        """
        return circle_to_compas(self.geometry)



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import Rhino
from Rhino.Geometry import Point3d, Vector3d, Plane
from Rhino.Geometry.Intersect.Intersection import MeshPlane

from compas.plugins import plugin


__all__ = [
    'trimesh_slice',
]


@plugin(category="trimesh", requires=['Rhino'])
def trimesh_slice(mesh, planes):
    """Slice a mesh by a list of planes.

    Parameters
    ----------
    mesh : tuple[sequence[[float, float, float] or :class:`compas.geometry.Point`], sequence[[int, int, int]]]
        The mesh to slice.
    planes : sequence[[point, vector] or :class:`compas.geometry.Plane`]
        The slicing planes.

    Returns
    -------
    list[list[[float, float, float]]]
        The points defining the slice polylines.

    Examples
    --------
    >>> from compas.geometry import Sphere, Plane, Point, Vector
    >>> sphere = Sphere([1, 1, 1], 1)
    >>> sphere = sphere.to_vertices_and_faces(u=30, v=30)
    >>> P1 = Plane(Point(0, 0, 0), Vector(0, -1, 0))
    >>> P2 = Plane(Point(0, 0, 0), Vector(0.87, -0.5, 0))
    >>> planes = [P1, P2]
    >>> points = trimesh_slice(sphere, planes)

    """
    # (0) see if input is already Rhino.Geometry.Mesh
    M = Rhino.Geometry.Mesh()
    if not isinstance(mesh, Rhino.Geometry.Mesh):
        for x, y, z in mesh[0]:
            M.Vertices.Add(x, y, z)
        for face in mesh[1]:
            M.Faces.AddFace(*face)
    else:
        M = mesh
    # (1) parse to Rhino.Geometry.Plane
    P = []
    for plane in planes:
        point = Point3d(plane[0][0], plane[0][1], plane[0][2])
        normal = Vector3d(plane[1][0], plane[1][1], plane[1][2])
        P.append(Plane(point, normal))
    # (2) Slice
    polylines = MeshPlane(M, P)
    # (3) Return points in a list of arrays
    polyline_pts = []
    for polyline in polylines:
        pts = []
        for i in range(polyline.Count):
            pts.append([polyline.X[i], polyline.Y[i], polyline.Z[i]])
        polyline_pts.append(pts)

    return polyline_pts



from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import bpy

from functools import partial

import compas_blender
from compas_blender.utilities import RGBColor
from compas.datastructures import Mesh
from compas.geometry import add_vectors
from compas.geometry import centroid_points
from compas.geometry import scale_vector

from compas.utilities import color_to_colordict
from compas.artists import MeshArtist
from .artist import BlenderArtist

colordict = partial(color_to_colordict, colorformat='rgb', normalize=True)


class MeshArtist(BlenderArtist, MeshArtist):
    """Artist for drawing mesh data structures in Blender.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        A COMPAS mesh.
    collection : str or :blender:`bpy.types.Collection`
        The Blender scene collection the object(s) created by this artist belong to.
    vertices : list[int], optional
        A list of vertex identifiers.
        Default is None, in which case all vertices are drawn.
    edges : list[tuple[int, int]], optional
        A list of edge keys (as uv pairs) identifying which edges to draw.
        The default is None, in which case all edges are drawn.
    faces : list[int], optional
        A list of face identifiers.
        The default is None, in which case all faces are drawn.
    vertexcolor : rgb-tuple or dict[int, rgb-tuple], optional
        The color specification for the vertices.
    edgecolor : rgb-tuple or dict[tuple[int, int], rgb-tuple], optional
        The color specification for the edges.
    facecolor : rgb-tuple or dict[int, rgb-tuple], optional
        The color specification for the faces.
    show_mesh : bool, optional
        If True, draw the mesh.
    show_vertices : bool, optional
        If True, draw the individual vertices.
    show_edges : bool, optional
        If True, draw the individual edges.
    show_faces : bool, optional
        If True, draw the individual faces.

    Attributes
    ----------
    vertexcollection : :blender:`bpy.types.Collection`
        The collection containing the vertices.
    edgecollection : :blender:`bpy.types.Collection`
        The collection containing the edges.
    facecollection : :blender:`bpy.types.Collection`
        The collection containing the faces.
    vertexlabelcollection : :blender:`bpy.types.Collection`
        The collection containing the vertex labels.
    edgelabelcollection : :blender:`bpy.types.Collection`
        The collection containing the edge labels.
    facelabelcollection : :blender:`bpy.types.Collection`
        The collection containing the face labels.

    Examples
    --------
    Use the Blender artist explicitly.

    .. code-block:: python

        from compas.datastructures import Mesh
        from compas_blender.artists import MeshArtist

        mesh = Mesh.from_meshgrid(10, 10)

        artist = MeshArtist(mesh)
        artist.draw()

    Or, use the artist through the plugin mechanism.

    .. code-block:: python

        from compas.datastructures import Mesh
        from compas.artists import Artist

        mesh = Mesh.from_meshgrid(10, 10)

        artist = Artist(mesh)
        artist.draw()

    """

    def __init__(self,
                 mesh: Mesh,
                 collection: Optional[Union[str, bpy.types.Collection]] = None,
                 vertices: Optional[List[int]] = None,
                 edges: Optional[List[int]] = None,
                 faces: Optional[List[int]] = None,
                 vertexcolor: Optional[RGBColor] = None,
                 edgecolor: Optional[RGBColor] = None,
                 facecolor: Optional[RGBColor] = None,
                 show_mesh: bool = False,
                 show_vertices: bool = True,
                 show_edges: bool = True,
                 show_faces: bool = True,
                 **kwargs: Any):

        super().__init__(mesh=mesh, collection=collection or mesh.name, **kwargs)

        self.vertices = vertices
        self.edges = edges
        self.faces = faces
        self.vertex_color = vertexcolor
        self.edge_color = edgecolor
        self.face_color = facecolor
        self.show_mesh = show_mesh
        self.show_vertices = show_vertices
        self.show_edges = show_edges
        self.show_faces = show_faces

    @property
    def vertexcollection(self) -> bpy.types.Collection:
        if not self._vertexcollection:
            self._vertexcollection = compas_blender.create_collection('Vertices', parent=self.collection)
        return self._vertexcollection

    @property
    def edgecollection(self) -> bpy.types.Collection:
        if not self._edgecollection:
            self._edgecollection = compas_blender.create_collection('Edges', parent=self.collection)
        return self._edgecollection

    @property
    def facecollection(self) -> bpy.types.Collection:
        if not self._facecollection:
            self._facecollection = compas_blender.create_collection('Faces', parent=self.collection)
        return self._facecollection

    @property
    def vertexnormalcollection(self) -> bpy.types.Collection:
        if not self._vertexnormalcollection:
            self._vertexnormalcollection = compas_blender.create_collection('VertexNormals', parent=self.collection)
        return self._vertexnormalcollection

    @property
    def facenormalcollection(self) -> bpy.types.Collection:
        if not self._facenormalcollection:
            self._facenormalcollection = compas_blender.create_collection('FaceNormals', parent=self.collection)
        return self._facenormalcollection

    @property
    def vertexlabelcollection(self) -> bpy.types.Collection:
        if not self._vertexlabelcollection:
            self._vertexlabelcollection = compas_blender.create_collection('VertexLabels', parent=self.collection)
        return self._vertexlabelcollection

    @property
    def edgelabelcollection(self) -> bpy.types.Collection:
        if not self._edgelabelcollection:
            self._edgelabelcollection = compas_blender.create_collection('EdgeLabels', parent=self.collection)
        return self._edgelabelcollection

    @property
    def facelabelcollection(self) -> bpy.types.Collection:
        if not self._facelabelcollection:
            self._facelabelcollection = compas_blender.create_collection('FaceLabels', parent=self.collection)
        return self._facelabelcollection

    # ==========================================================================
    # clear
    # ==========================================================================

    def clear_vertices(self):
        """Clear the objects contained in the vertex collection (``self.vertexcollection``).

        Returns
        -------
        None

        """
        compas_blender.delete_objects(self.vertexcollection.objects)

    def clear_edges(self):
        """Clear the objects contained in the edge collection (``self.edgecollection``).

        Returns
        -------
        None

        """
        compas_blender.delete_objects(self.edgecollection.objects)

    def clear_faces(self):
        """Clear the objects contained in the face collection (``self.facecollection``).

        Returns
        -------
        None

        """
        compas_blender.delete_objects(self.facecollection.objects)

    # ==========================================================================
    # draw
    # ==========================================================================

    def draw(self,
             vertices: Optional[List[int]] = None,
             edges: Optional[List[Tuple[int, int]]] = None,
             faces: Optional[List[int]] = None,
             vertexcolor: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None,
             edgecolor: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None,
             facecolor: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
             ) -> None:
        """Draw the mesh using the chosen visualization settings.

        Parameters
        ----------
        vertices : list[int], optional
            A list of vertex keys identifying which vertices to draw.
            Default is None, in which case all vertices are drawn.
        edges : list[tuple[int, int]]
            A list of edge keys (as uv pairs) identifying which edges to draw.
            The default is None, in which case all edges are drawn.
        faces : list[int]
            A list of face keys identifying which faces to draw.
            The default is None, in which case all faces are drawn.
        vertexcolor : rgb-tuple or dict[int, rgb-tuple]
            The color specification for the vertices.
        edgecolor : rgb-tuple or dict[tuple[int, int], rgb-tuple]
            The color specification for the edges.
        facecolor : rgb-tuple or dict[int, rgb-tuple]
            The color specification for the faces.

        Returns
        -------
        None

        """
        self.clear()
        if self.show_mesh:
            self.draw_mesh()
        if self.show_vertices:
            self.draw_vertices(vertices=vertices, color=vertexcolor)
        if self.show_edges:
            self.draw_edges(edges=edges, color=edgecolor)
        if self.show_faces:
            self.draw_faces(faces=faces, color=facecolor)

    def draw_mesh(self) -> List[bpy.types.Object]:
        """Draw the mesh.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        vertices, faces = self.mesh.to_vertices_and_faces()
        obj = compas_blender.draw_mesh(vertices, faces, name=self.mesh.name, collection=self.collection)
        return [obj]

    def draw_vertices(self,
                      vertices: Optional[List[int]] = None,
                      color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
                      ) -> List[bpy.types.Object]:
        """Draw a selection of vertices.

        Parameters
        ----------
        vertices : list[int], optional
            A list of vertex keys identifying which vertices to draw.
            Default is None, in which case all vertices are drawn.
        color : rgb-tuple or dict[int, rgb-tuple], optional
            The color specification for the vertices.
            The default color of vertices is :attr:`default_vertexcolor`.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        self.vertex_color = color
        vertices = vertices or self.vertices
        points = []
        for vertex in vertices:
            points.append({
                'pos': self.vertex_xyz[vertex],
                'name': f"{self.mesh.name}.vertex.{vertex}",
                'color': self.vertex_color.get(vertex, self.default_vertexcolor),
                'radius': 0.01
            })
        return compas_blender.draw_points(points, self.vertexcollection)

    def draw_edges(self,
                   edges: Optional[List[Tuple[int, int]]] = None,
                   color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
                   ) -> List[bpy.types.Object]:
        """Draw a selection of edges.

        Parameters
        ----------
        edges : list[tuple[int, int]], optional
            A list of edge keys (as uv pairs) identifying which edges to draw.
            The default is None, in which case all edges are drawn.
        color : rgb-tuple or dict[tuple[int, int], rgb-tuple], optional
            The color specification for the edges.
            The default color of edges is :attr:`default_edgecolor`.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        self.edge_color = color
        edges = edges or self.edges
        lines = []
        for edge in edges:
            lines.append({
                'start': self.vertex_xyz[edge[0]],
                'end': self.vertex_xyz[edge[1]],
                'color': self.edge_color.get(edge, self.default_edgecolor),
                'name': f"{self.mesh.name}.edge.{edge[0]}-{edge[1]}"
            })
        return compas_blender.draw_lines(lines, self.edgecollection)

    def draw_faces(self,
                   faces: Optional[List[int]] = None,
                   color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
                   ) -> List[bpy.types.Object]:
        """Draw a selection of faces.

        Parameters
        ----------
        faces : list[int], optional
            A list of face keys identifying which faces to draw.
            The default is None, in which case all faces are drawn.
        color : rgb-tuple or dict[int, rgb-tuple], optional
            The color specification for the faces.
            Th default color of faces is :attr:`default_facecolor`.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        self.face_color = color
        faces = faces or self.faces
        facets = []
        for face in faces:
            facets.append({
                'points': [self.vertex_xyz[vertex] for vertex in self.mesh.face_vertices(face)],
                'name': f"{self.mesh.name}.face.{face}",
                'color': self.face_color.get(face, self.default_facecolor)
            })
        return compas_blender.draw_faces(facets, self.facecollection)

    # ==========================================================================
    # draw normals
    # ==========================================================================

    def draw_vertexnormals(self,
                           vertices: Optional[List[int]] = None,
                           color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = (0.0, 1.0, 0.0),
                           scale: float = 1.0) -> List[bpy.types.Object]:
        """Draw the normals at the vertices of the mesh.

        Parameters
        ----------
        vertices : list[int], optional
            A selection of vertex normals to draw.
            Default is to draw all vertex normals.
        color : rgb-tuple or dict[int, rgb-tuple], optional
            The color specification of the normal vectors.
        scale : float, optional
            Scale factor for the vertex normals.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        vertices = vertices or self.vertices
        vertex_color = colordict(color, vertices, default=color)
        lines = []
        for vertex in vertices:
            a = self.vertex_xyz[vertex]
            n = self.mesh.vertex_normal(vertex)
            b = add_vectors(a, scale_vector(n, scale))
            lines.append({
                'start': a,
                'end': b,
                'color': vertex_color[vertex],
                'name': "{}.vertexnormal.{}".format(self.mesh.name, vertex)
            })
        return compas_blender.draw_lines(lines, collection=self.vertexnormalcollection)

    def draw_facenormals(self,
                         faces: Optional[List[List[int]]] = None,
                         color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = (0.0, 1.0, 1.0),
                         scale: float = 1.0) -> List[bpy.types.Object]:
        """Draw the normals of the faces.

        Parameters
        ----------
        faces : list[int], optional
            A selection of face normals to draw.
            Default is to draw all face normals.
        color : rgb-tuple or dict[int, rgb-tuple], optional
            The color specification of the normal vectors.
        scale : float, optional
            Scale factor for the face normals.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        faces = faces or self.faces
        face_color = colordict(color, faces, default=color)
        lines = []
        for face in faces:
            a = centroid_points([self.vertex_xyz[vertex] for vertex in self.mesh.face_vertices(face)])
            n = self.mesh.face_normal(face)
            b = add_vectors(a, scale_vector(n, scale))
            lines.append({
                'start': a,
                'end': b,
                'name': "{}.facenormal.{}".format(self.mesh.name, face),
                'color': face_color[face]
            })
        return compas_blender.draw_lines(lines, collection=self.facenormalcollection)

    # ==========================================================================
    # draw labels
    # ==========================================================================

    def draw_vertexlabels(self,
                          text: Optional[Dict[int, str]] = None,
                          color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
                          ) -> List[bpy.types.Object]:
        """Draw labels for a selection vertices.

        Parameters
        ----------
        text : dict[int, str], optional
            A dictionary of vertex labels as vertex-text pairs.
            The default value is None, in which case every vertex will be labeled with its identifier.
        color : rgb-tuple or dict[int, rgb-tuple], optional
            The color specification of the labels.
            The default color is the same as the default vertex color, :attr:`default_vertexcolor`.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        if not text or text == 'key':
            vertex_text = {vertex: str(vertex) for vertex in self.vertices}
        elif text == 'index':
            vertex_text = {vertex: str(index) for index, vertex in enumerate(self.vertices)}
        elif isinstance(text, dict):
            vertex_text = text
        else:
            raise NotImplementedError
        vertex_color = colordict(color, vertex_text, default=self.default_vertexcolor)
        labels = []
        for vertex in vertex_text:
            labels.append({
                'pos': self.vertex_xyz[vertex],
                'name': "{}.vertexlabel.{}".format(self.mesh.name, vertex),
                'text': vertex_text[vertex],
                'color': vertex_color[vertex]
            })
        return compas_blender.draw_texts(labels, collection=self.vertexlabelcollection)

    def draw_edgelabels(self,
                        text: Optional[Dict[Tuple[int, int], str]] = None,
                        color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
                        ) -> List[bpy.types.Object]:
        """Draw labels for a selection of edges.

        Parameters
        ----------
        text : dict[tuple[int, int], str], optional
            A dictionary of edge labels as edge-text pairs.
            The default value is None, in which case every edge will be labeled with its identifier.
        color : rgb-tuple or dict[tuple[int, int], rgb-tuple], optional
            The color specification of the labels.
            The default color is the same as the default color for edges, :attr:`default_edgecolor`.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        if text is None:
            edge_text = {(u, v): "{}-{}".format(u, v) for u, v in self.edges}
        elif isinstance(text, dict):
            edge_text = text
        else:
            raise NotImplementedError
        edge_color = colordict(color, edge_text, default=self.default_edgecolor)
        labels = []
        for edge in edge_text:
            labels.append({
                'pos': centroid_points([self.vertex_xyz[edge[0]], self.vertex_xyz[edge[1]]]),
                'name': "{}.edgelabel.{}-{}".format(self.mesh.name, *edge),
                'text': edge_text[edge]
            })
        return compas_blender.draw_texts(labels, collection=self.edgelabelcollection, color=edge_color)

    def draw_facelabels(self,
                        text: Optional[Dict[int, str]] = None,
                        color: Optional[Union[str, RGBColor, List[RGBColor], Dict[int, RGBColor]]] = None
                        ) -> List[bpy.types.Object]:
        """Draw labels for a selection of faces.

        Parameters
        ----------
        text : dict[int, str], optional
            A dictionary of face labels as face-text pairs.
            The default value is None, in which case every face will be labeled with its identifier.
        color : rgb-tuple or dict[int, rgb-tuple], optional
            The color specification of the labels.
            The default color is the same as the default face color: :attr:`default_facecolor`.

        Returns
        -------
        list[:blender:`bpy.types.Object`]

        """
        if not text or text == 'key':
            face_text = {face: str(face) for face in self.faces}
        elif text == 'index':
            face_text = {face: str(index) for index, face in enumerate(self.faces)}
        elif isinstance(text, dict):
            face_text = text
        else:
            raise NotImplementedError
        face_color = color or self.default_facecolor
        labels = []
        for face in face_text:
            labels.append({
                'pos': centroid_points([self.vertex_xyz[vertex] for vertex in self.mesh.face_vertices(face)]),
                'name': "{}.facelabel.{}".format(self.mesh.name, face),
                'text': face_text[face]
            })
        return compas_blender.draw_texts(labels, collection=self.collection, color=face_color)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from .assembly import Assembly  # noqa: F401 F403
from .part import Part  # noqa: F401 F403



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function


from compas.geometry import oriented_bounding_box_numpy
from compas.geometry import oriented_bounding_box_xy_numpy


__all__ = [
    'mesh_oriented_bounding_box_numpy',
    'mesh_oriented_bounding_box_xy_numpy',
]


def mesh_oriented_bounding_box_numpy(mesh):
    """Compute the (axis aligned) bounding box of a mesh.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        The mesh data structure.

    Returns
    -------
    list[list[float]]
        The bounding box of the mesh as a list of corner vertex coordinates.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> hypar = Mesh.from_obj(compas.get('hypar.obj'))
    >>> box = mesh_oriented_bounding_box_numpy(hypar)
    >>> len(box)
    8

    """
    xyz = mesh.vertices_attributes('xyz')
    return oriented_bounding_box_numpy(xyz)


def mesh_oriented_bounding_box_xy_numpy(mesh):
    """Compute the (axis aligned) bounding box of a projection of the mesh in the XY plane.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        The mesh data structure.

    Returns
    -------
    list[list[float]]
        The bounding box.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_obj(compas.get('faces.obj'))
    >>> mesh_oriented_bounding_box_xy_numpy(mesh)
    [[10.0, 0.0], [0.0, 0.0], [0.0, 10.0], [10.0, 10.0]]

    """
    xyz = mesh.vertices_attributes('xyz')
    return oriented_bounding_box_xy_numpy(xyz)



import bpy


__all__ = [
    "delete_unused_data",
]


def delete_unused_data():
    """Delete all collections, mesh and curve objects, meshes, curves, materials.

    Returns
    -------
    None

    """
    for block in bpy.data.meshes:
        if block.users == 0:
            bpy.data.meshes.remove(block)

    for block in bpy.data.curves:
        if block.users == 0:
            bpy.data.curves.remove(block)

    for block in bpy.data.materials:
        if block.users == 0:
            bpy.data.materials.remove(block)

    for block in bpy.data.textures:
        if block.users == 0:
            bpy.data.textures.remove(block)

    for block in bpy.data.images:
        if block.users == 0:
            bpy.data.images.remove(block)

    # for collection in bpy.context.scene.collection.children:
    #     bpy.context.scene.collection.children.unlink(collection)

    # for block in bpy.data.collections:
    #     objects = [o for o in block.objects if o.users]
    #     while objects:
    #         bpy.data.objects.remove(objects.pop())
    #     for collection in block.children:
    #         block.children.unlink(collection)
    #     if block.users == 0:
    #         bpy.data.collections.remove(block)



from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from compas.geometry import cross_vectors
from compas.geometry import subtract_vectors
from compas.geometry import dot_vectors
from compas.geometry import cross_vectors_xy


__all__ = [
    'convex_hull',
    'convex_hull_xy',
]


def convex_hull(points):
    """Construct convex hull for a set of points.

    Parameters
    ----------
    points : sequence[point]
        A sequence of XYZ coordinates.

    Returns
    -------
    list[[int, int, int]]
        The triangular faces of the convex hull as lists of vertex indices
        referring to the original point coordinates.

    Notes
    -----
    This algorithm is based on [1]_. Note that is not optimized and relatively
    slow on large sets of points. For a more optimized version of this algorithm,
    see [2]_.

    References
    ----------
    .. [1] GitHubGist. *Convex Hull*.
           Available at: https://gist.github.com/anonymous/5184ba0bcab21d3dd19781efd3aae543
    .. [2] Thomas Diewald. *Convex Hull 3D - Quickhull Algorithm*.
           Available at: https://web.archive.org/web/20180106161310/http://thomasdiewald.com/blog/?p=1888

    Examples
    --------
    >>>

    """
    def _normal_face(face):
        u = subtract_vectors(points[face[1]], points[face[0]])
        v = subtract_vectors(points[face[-1]], points[face[0]])
        return cross_vectors(u, v)

    def _seen(face, p):
        normal = _normal_face(face)
        vec = subtract_vectors(points[p], points[face[0]])
        return (dot_vectors(normal, vec) >= 0)

    def _bdry(faces):
        bdry_fw = set([(face[i - 1], face[i]) for face in faces for i in range(len(face))])
        bdry_bk = set([(face[i], face[i - 1]) for face in faces for i in range(len(face))])
        return bdry_fw - bdry_bk

    def _add_point(hull, p):
        seen_faces = [face for face in hull if _seen(face, p)]

        if len(seen_faces) == len(hull):
            # if can see all faces, unsee ones looking "down"
            normal = _normal_face(seen_faces[0])
            seen_faces = [face for face in seen_faces if dot_vectors(_normal_face(face), normal) > 0]

        for face in seen_faces:
            hull.remove(face)

        for edge in _bdry(seen_faces):
            hull.append([edge[0], edge[1], p])

    hull = [[0, 1, 2], [0, 2, 1]]
    for i in range(3, len(points)):
        _add_point(hull, i)
    return hull


def convex_hull_xy(points, strict=False):
    """Computes the convex hull of a set of 2D points.

    Parameters
    ----------
    points : sequence[point]
        XY(Z) coordinates of the points.

    Returns
    -------
    list[[float, float, 0.0]]
        XY(Z) coordinates of vertices of the convex hull in counter-clockwise order,
        starting from the vertex with the lexicographically smallest coordinates.

    Notes
    -----
    Implements Andrew's monotone chain algorithm [1]_.
    O(n log n) complexity.

    References
    ----------
    .. [1] Wiki Books. *Algorithm Implementation/Geometry/Convex hull/Monotone chain*.
           Available at: https://en.wikibooks.org/wiki/Algorithm_Implementation/Geometry/Convex_hull/Monotone_chain.

    Examples
    --------
    >>>

    """
    def cross(o, a, b):
        u = subtract_vectors(a, o)
        v = subtract_vectors(b, o)
        return cross_vectors_xy(u, v)[2]

    # Sort the points lexicographically (tuples are compared lexicographically).
    # Remove duplicates to detect the case we have just one unique point.
    points = sorted(set(map(tuple, points)))

    # Boring case: no points or a single point, possibly repeated multiple times.
    if len(points) <= 1:
        return points

    # Build lower hull
    lower = []
    for p in points:
        if strict:
            while len(lower) >= 2 and cross(lower[-2], lower[-1], p) < 0:
                lower.pop()
        else:
            while len(lower) >= 2 and cross(lower[-2], lower[-1], p) <= 0:
                lower.pop()
        lower.append(p)

    # Build upper hull
    upper = []
    for p in reversed(points):
        if strict:
            while len(upper) >= 2 and cross(upper[-2], upper[-1], p) < 0:
                upper.pop()
        else:
            while len(upper) >= 2 and cross(upper[-2], upper[-1], p) <= 0:
                upper.pop()
        upper.append(p)

    # Concatenation of the lower and upper hulls gives the convex hull.
    # Last point of each list is omitted because it is repeated at the beginning of the other list.
    return lower[:-1] + upper[:-1]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from compas.data import Data
from compas.files import URDFElement
from compas.files import URDFParser
from compas.geometry import Box
from compas.geometry import Capsule
from compas.geometry import Cylinder
from compas.geometry import Frame
from compas.geometry import Sphere
from compas.geometry import Transformation
from compas.robots.model.base import FrameProxy
from compas.robots.model.base import _attr_from_data
from compas.robots.model.base import _attr_to_data
from compas.robots.model.geometry import BoxProxy
from compas.robots.model.geometry import CapsuleProxy
from compas.robots.model.geometry import Color
from compas.robots.model.geometry import CylinderProxy
from compas.robots.model.geometry import Geometry
from compas.robots.model.geometry import Material
from compas.robots.model.geometry import MeshDescriptor
from compas.robots.model.geometry import SphereProxy
from compas.robots.model.geometry import Texture

__all__ = ['Link', 'Inertial', 'Visual', 'Collision', 'Mass', 'Inertia']


class Mass(Data):
    """Represents a value of mass usually related to a link."""

    def __init__(self, value):
        super(Mass, self).__init__()
        self.value = float(value)

    def __str__(self):
        return str(self.value)

    def get_urdf_element(self):
        attributes = {'value': self.value}
        return URDFElement('mass', attributes)

    @property
    def data(self):
        return {'value': self.value}

    @data.setter
    def data(self, data):
        self.value = data['value']

    @classmethod
    def from_data(cls, data):
        return cls(**data)


class Inertia(Data):
    """Rotational inertia matrix (3x3) represented in the inertia frame.

    Since the rotational inertia matrix is symmetric, only 6 above-diagonal
    elements of this matrix are specified here, using the attributes
    ``ixx``, ``ixy``, ``ixz``, ``iyy``, ``iyz``, ``izz``.

    """

    def __init__(self, ixx=0., ixy=0., ixz=0., iyy=0., iyz=0., izz=0.):
        super(Inertia, self).__init__()
        self.ixx = float(ixx)
        self.ixy = float(ixy)
        self.ixz = float(ixz)
        self.iyy = float(iyy)
        self.iyz = float(iyz)
        self.izz = float(izz)

    def get_urdf_element(self):
        attributes = {
            'ixx': self.ixx,
            'ixy': self.ixy,
            'ixz': self.ixz,
            'iyy': self.iyy,
            'iyz': self.iyz,
            'izz': self.izz,
        }
        return URDFElement('inertia', attributes)

    @property
    def data(self):
        return {
            'ixx': self.ixx,
            'ixy': self.ixy,
            'ixz': self.ixz,
            'iyy': self.iyy,
            'iyz': self.iyz,
            'izz': self.izz,
        }

    @data.setter
    def data(self, data):
        self.ixx = data.get('ixx', 0.)
        self.ixy = data.get('ixy', 0.)
        self.ixz = data.get('ixz', 0.)
        self.iyy = data.get('iyy', 0.)
        self.iyz = data.get('iyz', 0.)
        self.izz = data.get('izz', 0.)

    @classmethod
    def from_data(cls, data):
        return cls(**data)


class Inertial(Data):
    """Inertial properties of a link.

    Attributes
    ----------
    origin
        This is the pose of the inertial reference frame,
        relative to the link reference frame.
    mass
        Mass of the link.
    inertia
        3x3 rotational inertia matrix, represented in the inertia frame.

    """

    def __init__(self, origin=None, mass=None, inertia=None):
        super(Inertial, self).__init__()
        self.origin = origin
        self.mass = mass
        self.inertia = inertia

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = FrameProxy.create_proxy(value)

    def get_urdf_element(self):
        elements = [self.origin, self.mass, self.inertia]
        return URDFElement('inertial', elements=elements)

    @property
    def data(self):
        return {
            'origin': self.origin.data if self.origin else None,
            'mass': self.mass.data if self.mass else None,
            'inertia': self.inertia.data if self.inertia else None,
        }

    @data.setter
    def data(self, data):
        self.origin = Frame.from_data(data['origin']) if data['origin'] else None
        self.mass = Mass.from_data(data['mass']) if data['mass'] else None
        self.inertia = Inertia.from_data(data['inertia']) if data['inertia'] else None


class LinkItem(object):
    def __init__(self):
        self.init_transformation = None  # to store the init transformation
        self.current_transformation = None  # to store the current transformation
        self.native_geometry = None  # to store the link's CAD native geometry


class Visual(LinkItem, Data):
    """Visual description of a link.

    Attributes
    ----------
    geometry
        Shape of the visual element.
    origin
        Reference frame of the visual element with respect
        to the reference frame of the link.
    name
        Name of the visual element.
    material
        Material of the visual element.
    attr
        Non-standard attributes.

    """

    def __init__(self, geometry, origin=None, name=None, material=None, **kwargs):
        super(Visual, self).__init__()
        self.geometry = geometry
        self.origin = origin
        self.name = name
        self.material = material
        self.attr = kwargs

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = FrameProxy.create_proxy(value)

    def get_urdf_element(self):
        attributes = {}
        if self.name is not None:
            attributes['name'] = self.name
        attributes.update(self.attr)
        elements = [self.origin, self.geometry, self.material]
        return URDFElement('visual', attributes, elements)

    # Overriding the default name property, because sometimes the name really is `None`.
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def data(self):
        return {
            'geometry': self.geometry.data,
            'origin': self.origin.data if self.origin else None,
            'name': self.name,
            'material': self.material.data if self.material else None,
            'attr': _attr_to_data(self.attr),
            'init_transformation': self.init_transformation.data if self.init_transformation else None,
            'current_transformation': self.current_transformation.data if self.current_transformation else None,
        }

    @data.setter
    def data(self, data):
        self.geometry = Geometry.from_data(data['geometry'])
        self.origin = Frame.from_data(data['origin']) if data['origin'] else None
        self.name = data['name']
        self.material = Material.from_data(data['material']) if data['material'] else None
        self.attr = _attr_from_data(data['attr'])
        self.init_transformation = Transformation.from_data(data['init_transformation']) if data['init_transformation'] else None
        self.current_transformation = Transformation.from_data(data['current_transformation']) if data['current_transformation'] else None

    @classmethod
    def from_data(cls, data):
        visual = cls(Geometry.from_data(data['geometry']))
        visual.data = data
        return visual

    def get_color(self):
        """Get the RGBA color array assigned to the link.

        Only if the link has a material assigned.

        Returns
        -------
        list[float]
            List of 4 floats (``0.0-1.0``) indicating RGB colors and Alpha channel.

        """
        if self.material:
            return self.material.get_color()
        else:
            return None

    @classmethod
    def from_primitive(cls, primitive, **kwargs):
        """Create visual link from a primitive shape.

        Parameters
        ----------
        primitive : :compas:`compas.geometry.Shape`
            A primitive shape.
        **kwargs : dict[str, Any], optional
            The keyword arguments (kwargs) collected in a dict.
            These allow using non-standard attributes absent in the URDF specification.

        Returns
        -------
        :class:`compas.datastructures.Mesh`
            A visual description object.
        """
        geometry = Geometry()
        geometry.shape = primitive
        return cls(geometry, **kwargs)


class Collision(LinkItem, Data):
    """Collidable description of a link.

    Attributes
    ----------
    geometry
        Shape of the collidable element.
    origin
        Reference frame of the collidable element with respect
        to the reference frame of the link.
    name
        Name of the collidable element.
    attr
        Non-standard attributes.

    """

    def __init__(self, geometry, origin=None, name=None, **kwargs):
        super(Collision, self).__init__()
        self.geometry = geometry
        self.origin = origin
        self.name = name
        self.attr = kwargs

    @property
    def origin(self):
        return self._origin

    @origin.setter
    def origin(self, value):
        self._origin = FrameProxy.create_proxy(value)

    def get_urdf_element(self):
        attributes = {}
        if self.name is not None:
            attributes['name'] = self.name
        attributes.update(self.attr)
        elements = [self.origin, self.geometry]
        return URDFElement('collision', attributes, elements)

    # Overriding the default name property, because sometimes the name really is `None`.
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        self._name = name

    @property
    def data(self):
        return {
            'geometry': self.geometry.data,
            'origin': self.origin.data if self.origin else None,
            'name': self.name,
            'attr': _attr_to_data(self.attr),
            'init_transformation': self.init_transformation.data if self.init_transformation else None,
            'current_transformation': self.current_transformation.data if self.current_transformation else None,
        }

    @data.setter
    def data(self, data):
        self.geometry = Geometry.from_data(data['geometry'])
        self.origin = Frame.from_data(data['origin']) if data['origin'] else None
        self.name = data['name']
        self.attr = _attr_from_data(data['attr'])
        self.init_transformation = Transformation.from_data(data['init_transformation']) if data['init_transformation'] else None
        self.current_transformation = Transformation.from_data(data['current_transformation']) if data['current_transformation'] else None

    @classmethod
    def from_data(cls, data):
        collision = cls(Geometry.from_data(data['geometry']))
        collision.data = data
        return collision

    @classmethod
    def from_primitive(cls, primitive, **kwargs):
        """Create collision link from a primitive shape.

        Parameters
        ----------
        primitive : :compas:`compas.geometry.Shape`
            A primitive shape.
        **kwargs : dict[str, Any], optional
            The keyword arguments (kwargs) collected in a dict.
            These allow using non-standard attributes absent in the URDF specification.

        Returns
        -------
        :class:`compas.datastructures.Mesh`
            A collision description object.
        """
        geometry = Geometry()
        geometry.shape = primitive
        return cls(geometry, **kwargs)


class Link(Data):
    """Link represented as a rigid body with an inertia, visual, and collision features.

    Attributes
    ----------
    name
        Name of the link itself.
    type
        Link type. Undocumented in URDF, but used by PR2.
    visual
        Visual properties of the link.
    collision
        Collision properties of the link. This can be different
        from the visual properties of a link.
    inertial
        Inertial properties of the link.
    attr
        Non-standard attributes.
    joints
        A list of joints that are the link's children
    parent_joint
        The reference to a parent joint if it exists

    """

    def __init__(self, name, type=None, visual=(), collision=(), inertial=None, **kwargs):
        super(Link, self).__init__()
        self.name = name
        self.type = type
        self.visual = list(visual or [])
        self.collision = list(collision or [])
        self.inertial = inertial
        self.attr = kwargs
        self.joints = []
        self.parent_joint = None

    def get_urdf_element(self):
        attributes = {'name': self.name}
        if self.type is not None:
            attributes['type'] = self.type
        attributes.update(self.attr)
        elements = self.visual + self.collision + [self.inertial]
        return URDFElement('link', attributes, elements)

    @property
    def data(self):
        return {
            'name': self.name,
            'type': self.type,
            'visual': [visual.data for visual in self.visual],
            'collision': [collision.data for collision in self.collision],
            'inertial': self.inertial.data if self.inertial else None,
            'attr': _attr_to_data(self.attr),
            'joints': [joint.data for joint in self.joints],
        }

    @data.setter
    def data(self, data):
        from compas.robots.model.joint import Joint
        self.name = data['name']
        self.type = data['type']
        self.visual = [Visual.from_data(d) for d in data['visual']]
        self.collision = [Collision.from_data(d) for d in data['collision']]
        self.inertial = Inertial.from_data(data['inertial']) if data['inertial'] else None
        self.attr = _attr_from_data(data['attr'])
        self.joints = [Joint.from_data(d) for d in data['joints']]

    @classmethod
    def from_data(cls, data):
        link = cls(data['name'])
        link.data = data
        return link


URDFParser.install_parser(Link, 'robot/link')
URDFParser.install_parser(Inertial, 'robot/link/inertial')
URDFParser.install_parser(Mass, 'robot/link/inertial/mass')
URDFParser.install_parser(Inertia, 'robot/link/inertial/inertia')

URDFParser.install_parser(Visual, 'robot/link/visual')
URDFParser.install_parser(Collision, 'robot/link/collision')

URDFParser.install_parser(Frame, 'robot/link/inertial/origin', 'robot/link/visual/origin', 'robot/link/collision/origin', proxy_type=FrameProxy)
URDFParser.install_parser(Geometry, 'robot/link/visual/geometry', 'robot/link/collision/geometry')
URDFParser.install_parser(MeshDescriptor, 'robot/link/visual/geometry/mesh', 'robot/link/collision/geometry/mesh')
URDFParser.install_parser(Box, 'robot/link/visual/geometry/box', 'robot/link/collision/geometry/box', proxy_type=BoxProxy)
URDFParser.install_parser(Cylinder, 'robot/link/visual/geometry/cylinder', 'robot/link/collision/geometry/cylinder', proxy_type=CylinderProxy)
URDFParser.install_parser(Sphere, 'robot/link/visual/geometry/sphere', 'robot/link/collision/geometry/sphere', proxy_type=SphereProxy)
URDFParser.install_parser(Capsule, 'robot/link/visual/geometry/capsule', 'robot/link/collision/geometry/capsule', proxy_type=CapsuleProxy)

URDFParser.install_parser(Material, 'robot/link/visual/material')
URDFParser.install_parser(Color, 'robot/link/visual/material/color')
URDFParser.install_parser(Texture, 'robot/link/visual/material/texture')



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from math import cos
from math import pi
from math import sin
from math import sqrt

from compas.utilities import pairwise

from compas.geometry import matrix_from_frame
from compas.geometry import transform_points
from compas.geometry import Circle
from compas.geometry import Frame
from compas.geometry import Plane

from ._shape import Shape


class Cone(Shape):
    """A cone is defined by a circle and a height.

    Parameters
    ----------
    circle : [plane, radius] | :class:`compas.geometry.Circle`
        The base circle of the cone.
    height : float
        The height of the cone.

    Attributes
    ----------
    plane : :class:`compas.geometry.Plane`
        The plane of the cone.
    circle : :class:`compas.geometry.Circle`
        The circle of the cone.
    center : :class:`compas.geometry.Point`
        The center of the cone.
    radius : float
        The radius of the cone.
    height : float
        The height of the cone.
    normal : :class:`compas.geometry.Vector`, read-only
        The normal of the cone.
    diameter : float, read-only
        The diameter of the cone.
    area : float, read-only
        The surface area of the cone.
    volume : float, read-only
        The volume of the cone.

    Examples
    --------
    >>> from compas.geometry import Plane
    >>> from compas.geometry import Cone
    >>> plane = Plane([0, 0, 0], [0, 0, 1])
    >>> circle = Circle(plane, 5)
    >>> cone = Cone(circle, 7)

    """

    __slots__ = ['_circle', '_height']

    def __init__(self, circle, height, **kwargs):
        super(Cone, self).__init__(**kwargs)
        self._circle = None
        self._height = None
        self.circle = circle
        self.height = height

    # ==========================================================================
    # data
    # ==========================================================================

    @property
    def DATASCHEMA(self):
        """:class:`schema.Schema` : Schema of the data representation."""
        import schema
        return schema.Schema({
            'circle': {
                'plane': Plane.DATASCHEMA.fget(None),
                'radius': schema.And(float, lambda x: x > 0)
            },
            'height': schema.And(float, lambda x: x > 0)
        })

    @property
    def JSONSCHEMANAME(self):
        """str : Name of the schema of the data representation in JSON format."""
        return 'cone'

    @property
    def data(self):
        """dict : Returns the data dictionary that represents the cone.
        """
        return {'circle': self.circle.data, 'height': self.height}

    @data.setter
    def data(self, data):
        self.circle = Circle.from_data(data['circle'])
        self.height = data['height']

    @classmethod
    def from_data(cls, data):
        """Construct a cone from its data representation.

        Parameters
        ----------
        data : dict
            The data dictionary.

        Returns
        -------
        :class:`compas.geometry.Cone`
            The constructed cone.

        Examples
        --------
        >>> from compas.geometry import Cone
        >>> from compas.geometry import Circle
        >>> from compas.geometry import Plane
        >>> data = {'circle': Circle(Plane.worldXY(), 5).data, 'height': 7.}
        >>> cone = Cone.from_data(data)

        """
        cone = cls(Circle.from_data(data['circle']), data['height'])
        return cone

    # ==========================================================================
    # properties
    # ==========================================================================

    @property
    def plane(self):
        return self.circle.plane

    @plane.setter
    def plane(self, plane):
        self.circle.plane = Plane(*plane)

    @property
    def circle(self):
        return self._circle

    @circle.setter
    def circle(self, circle):
        self._circle = Circle(*circle)

    @property
    def radius(self):
        return self.circle.radius

    @radius.setter
    def radius(self, radius):
        self.circle.radius = float(radius)

    @property
    def height(self):
        return self._height

    @height.setter
    def height(self, height):
        self._height = float(height)

    @property
    def normal(self):
        return self.plane.normal

    @property
    def diameter(self):
        return self.circle.diameter

    @property
    def center(self):
        return self.circle.center

    @center.setter
    def center(self, point):
        self.circle.center = point

    @property
    def area(self):
        r = self.circle.radius
        return pi * r * (r + sqrt(self.height**2 + r**2))

    @property
    def volume(self):
        return pi * self.circle.radius**2 * (self.height / 3)

    # ==========================================================================
    # customisation
    # ==========================================================================

    def __repr__(self):
        return 'Cone({0!r}, {1!r})'.format(self.circle, self.height)

    def __len__(self):
        return 2

    def __getitem__(self, key):
        if key == 0:
            return self.circle
        elif key == 1:
            return self.height
        else:
            raise KeyError

    def __setitem__(self, key, value):
        if key == 0:
            self.circle = value
        elif key == 1:
            self.height = value
        else:
            raise KeyError

    def __iter__(self):
        return iter([self.circle, self.height])

    # ==========================================================================
    # constructors
    # ==========================================================================

    # ==========================================================================
    # methods
    # ==========================================================================

    def to_vertices_and_faces(self, u=16, triangulated=False):
        """Returns a list of vertices and faces.

        Parameters
        ----------
        u : int, optional
            Number of faces in the "u" direction.
        triangulated: bool, optional
            If True, triangulate the faces.

        Returns
        -------
        list[list[float]]
            A list of vertex locations.
        list[list[int]]
            And a list of faces,
            with each face defined as a list of indices into the list of vertices.

        """
        if u < 3:
            raise ValueError('The value for u should be u > 3.')

        vertices = [[0, 0, 0]]
        a = 2 * pi / u
        radius = self.circle.radius
        for i in range(u):
            x = radius * cos(i * a)
            y = radius * sin(i * a)
            vertices.append([x, y, 0])
        vertices.append([0, 0, self.height])

        frame = Frame.from_plane(self.circle.plane)
        M = matrix_from_frame(frame)
        vertices = transform_points(vertices, M)

        faces = []
        first = 0
        last = len(vertices) - 1
        for i, j in pairwise(range(1, last)):
            faces.append([i, j, last])
            faces.append([j, i, first])
        faces.append([last - 1, 1, last])
        faces.append([1, last - 1, first])

        if triangulated:
            triangles = []
            for face in faces:
                if len(face) == 4:
                    triangles.append(face[0:3])
                    triangles.append([face[0], face[2], face[3]])
                else:
                    triangles.append(face)
            faces = triangles

        return vertices, faces

    def transform(self, transformation):
        """Transform the cone.

        Parameters
        ----------
        transformation : :class:`compas.geometry.Transformation`
            The transformation used to transform the cone.

        Returns
        -------
        None

        Examples
        --------
        >>> from compas.geometry import Frame
        >>> from compas.geometry import Transformation
        >>> from compas.geometry import Plane
        >>> from compas.geometry import Cone
        >>> from compas.geometry import Circle
        >>> circle = Circle(Plane.worldXY(), 5)
        >>> cone = Cone(circle, 7)
        >>> frame = Frame([1, 1, 1], [0.68, 0.68, 0.27], [-0.67, 0.73, -0.15])
        >>> T = Transformation.from_frame(frame)
        >>> cone.transform(T)

        """
        self.circle.transform(transformation)



"""
********************************************************************************
utilities
********************************************************************************

.. currentmodule:: compas_rhino.utilities


layers
======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    create_layers
    clear_layers
    delete_layers


objects
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    delete_object
    delete_objects
    get_line_coordinates
    get_objects
    get_object_attributes
    set_object_attributes
    get_object_attributes_from_name
    get_object_layers
    get_object_types
    get_object_names
    get_object_name
    get_point_coordinates
    get_polyline_coordinates
    get_polygon_coordinates
    purge_objects
    select_curve
    select_curves
    select_line
    select_lines
    select_mesh
    select_meshes
    select_point
    select_points
    select_polygon
    select_polygons
    select_polyline
    select_polylines
    select_surface
    select_surfaces


drawing
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    draw_labels
    draw_points
    draw_lines
    draw_polylines
    draw_faces
    draw_cylinders
    draw_pipes
    draw_spheres
    draw_mesh
    draw_curves
    draw_surfaces


constructors
============

.. autosummary::
    :toctree: generated/
    :nosignatures:

    volmesh_from_polysurfaces

"""
from __future__ import absolute_import

from .document import (
    get_document_basename,
    get_document_filename,
    get_document_extension,
    get_document_filepath,
    get_document_dirname
)
from .layers import (
    create_layers_from_path,
    create_layers_from_paths,
    create_layers_from_dict,
    create_layers,
    clear_layer,
    clear_current_layer,
    clear_layers,
    delete_layers
)
from .objects import (
    find_object,
    get_objects,
    get_object_layers,
    get_object_types,
    get_object_names,
    get_object_name,
    get_object_attributes,
    set_object_attributes,
    get_object_attributes_from_name,
    delete_object,
    delete_objects,
    purge_objects,
    is_curve_line,
    is_curve_polyline,
    is_curve_polygon,
    get_points,
    get_curves,
    get_lines,
    get_polylines,
    get_polygons,
    get_point_coordinates,
    get_line_coordinates,
    get_polyline_coordinates,
    get_polygon_coordinates,
    get_meshes,
    get_mesh_face_vertices,
    get_mesh_vertex_coordinates,
    get_mesh_vertex_colors,
    set_mesh_vertex_colors,
    get_mesh_vertices_and_faces,
    get_mesh_vertex_index,
    get_mesh_face_index,
    get_mesh_edge_index,
    select_object,
    select_objects,
    select_point,
    select_points,
    select_curve,
    select_curves,
    select_line,
    select_lines,
    select_polyline,
    select_polylines,
    select_polygon,
    select_polygons,
    select_surface,
    select_surfaces,
    select_mesh,
    select_meshes
)
from .misc import (
    wait,
    get_tolerance,
    toggle_toolbargroup,
    pick_point,
    browse_for_folder,
    browse_for_file,
    print_display_on,
    display_message,
    display_text,
    display_image,
    display_html,
    update_settings,
    update_named_values,
    screenshot_current_view,
    select_folder,
    select_file,
    unload_modules
)
from .drawing import (
    draw_labels,
    draw_points,
    draw_lines,
    draw_geodesics,
    draw_polylines,
    draw_breps,
    draw_faces,
    draw_cylinders,
    draw_pipes,
    draw_spheres,
    draw_mesh,
    draw_circles,
    draw_curves,
    draw_surfaces,
)
from .geometry import uv_points_from_surface
from .constructors import volmesh_from_polysurfaces


__all__ = [
    'get_document_basename',
    'get_document_filename',
    'get_document_extension',
    'get_document_filepath',
    'get_document_dirname',
    'create_layers_from_path',
    'create_layers_from_paths',
    'create_layers_from_dict',
    'create_layers',
    'clear_layer',
    'clear_current_layer',
    'clear_layers',
    'delete_layers',
    'find_object',
    'get_objects',
    'get_object_layers',
    'get_object_types',
    'get_object_names',
    'get_object_name',
    'get_object_attributes',
    'set_object_attributes',
    'get_object_attributes_from_name',
    'delete_object',
    'delete_objects',
    'purge_objects',
    'is_curve_line',
    'is_curve_polyline',
    'is_curve_polygon',
    'get_points',
    'get_curves',
    'get_lines',
    'get_polylines',
    'get_polygons',
    'get_point_coordinates',
    'get_line_coordinates',
    'get_polyline_coordinates',
    'get_polygon_coordinates',
    'get_meshes',
    'get_mesh_face_vertices',
    'get_mesh_vertex_coordinates',
    'get_mesh_vertex_colors',
    'set_mesh_vertex_colors',
    'get_mesh_vertices_and_faces',
    'get_mesh_vertex_index',
    'get_mesh_face_index',
    'get_mesh_edge_index',
    'select_object',
    'select_objects',
    'select_point',
    'select_points',
    'select_curve',
    'select_curves',
    'select_line',
    'select_lines',
    'select_polyline',
    'select_polylines',
    'select_polygon',
    'select_polygons',
    'select_surface',
    'select_surfaces',
    'select_mesh',
    'select_meshes',
    'wait',
    'get_tolerance',
    'toggle_toolbargroup',
    'pick_point',
    'browse_for_folder',
    'browse_for_file',
    'print_display_on',
    'display_message',
    'display_text',
    'display_image',
    'display_html',
    'update_settings',
    'update_named_values',
    'screenshot_current_view',
    'select_folder',
    'select_file',
    'unload_modules',
    'draw_labels',
    'draw_points',
    'draw_lines',
    'draw_geodesics',
    'draw_polylines',
    'draw_breps',
    'draw_faces',
    'draw_cylinders',
    'draw_pipes',
    'draw_spheres',
    'draw_mesh',
    'draw_circles',
    'draw_curves',
    'draw_surfaces',
    'uv_points_from_surface',
    'volmesh_from_polysurfaces'
]



"""
********************************************************************************
conversions
********************************************************************************

.. currentmodule:: compas_rhino.conversions

.. rst-class:: lead

Conversions between Rhino geometry objects (:mod:`Rhino.Geometry`) and COMPAS geometry objects (:mod:`compas.geometry`).

Classes
=======

.. autosummary::
    :toctree: generated/
    :nosignatures:

    RhinoGeometry
    RhinoBox
    RhinoCircle
    RhinoCone
    RhinoCurve
    RhinoCylinder
    RhinoEllipse
    RhinoLine
    RhinoMesh
    RhinoPlane
    RhinoPoint
    RhinoPolyline
    RhinoSphere
    RhinoSurface
    RhinoVector


Exceptions
==========

.. autosummary::
    :toctree: generated/
    :nosignatures:

    ConversionError


Functions
=========

Primitives
----------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    point_to_rhino
    vector_to_rhino
    line_to_rhino
    plane_to_rhino
    frame_to_rhino
    circle_to_rhino
    ellipse_to_rhino
    polyline_to_rhino
    polygon_to_rhino
    point_to_compas
    vector_to_compas
    line_to_compas
    plane_to_compas
    plane_to_compas_frame
    circle_to_compas
    ellipse_to_compas
    polyline_to_compas
    polygon_to_compas


Shapes
------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    box_to_rhino
    sphere_to_rhino
    cone_to_rhino
    cylinder_to_rhino
    box_to_compas
    sphere_to_compas
    cone_to_compas
    cylinder_to_compas


Curves
------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    line_to_rhino_curve
    circle_to_rhino_curve
    ellipse_to_rhino_curve
    curve_to_compas_line
    curve_to_compas_circle
    curve_to_compas_ellipse
    curve_to_compas_polyline


Surfaces
--------

.. autosummary::
    :toctree: generated/
    :nosignatures:


Transformations
---------------

.. autosummary::
    :toctree: generated/
    :nosignatures:

    xform_to_rhino

"""
from __future__ import absolute_import

from ._exceptions import ConversionError

from ._primitives import (
    point_to_rhino,
    vector_to_rhino,
    line_to_rhino,
    plane_to_rhino,
    frame_to_rhino,
    circle_to_rhino,
    ellipse_to_rhino,
    polyline_to_rhino,
    polygon_to_rhino,

    point_to_compas,
    vector_to_compas,
    line_to_compas,
    plane_to_compas,
    plane_to_compas_frame,
    circle_to_compas,
    ellipse_to_compas,
    polyline_to_compas,
    polygon_to_compas
)
from ._shapes import (
    box_to_rhino,
    sphere_to_rhino,
    cone_to_rhino,
    cylinder_to_rhino,

    box_to_compas,
    sphere_to_compas,
    cone_to_compas,
    cylinder_to_compas,
)
from ._curves import (
    line_to_rhino_curve,
    circle_to_rhino_curve,
    ellipse_to_rhino_curve,

    curve_to_compas_circle,
    curve_to_compas_ellipse,
    curve_to_compas_line,
    curve_to_compas_polyline
)
from ._transformations import (
    xform_to_rhino
)

from ._geometry import RhinoGeometry

from .box import RhinoBox
from .circle import RhinoCircle
from .cone import RhinoCone
from .curve import RhinoCurve
from .cylinder import RhinoCylinder
from .ellipse import RhinoEllipse
from .line import RhinoLine
from .mesh import RhinoMesh
from .plane import RhinoPlane
from .point import RhinoPoint
from .polyline import RhinoPolyline
from .sphere import RhinoSphere
from .surface import RhinoSurface
from .vector import RhinoVector

BaseRhinoGeometry = RhinoGeometry


__all__ = [
    'ConversionError',

    'point_to_rhino',
    'vector_to_rhino',
    'line_to_rhino',
    'plane_to_rhino',
    'frame_to_rhino',
    'circle_to_rhino',
    'ellipse_to_rhino',
    'polyline_to_rhino',
    'polygon_to_rhino',

    'point_to_compas',
    'vector_to_compas',
    'line_to_compas',
    'plane_to_compas',
    'plane_to_compas_frame',
    'circle_to_compas',
    'ellipse_to_compas',
    'polyline_to_compas',
    'polygon_to_compas',

    'box_to_rhino',
    'sphere_to_rhino',
    'cone_to_rhino',
    'cylinder_to_rhino',

    'box_to_compas',
    'sphere_to_compas',
    'cone_to_compas',
    'cylinder_to_compas',

    'line_to_rhino_curve',
    'circle_to_rhino_curve',
    'ellipse_to_rhino_curve',

    'curve_to_compas_circle',
    'curve_to_compas_ellipse',
    'curve_to_compas_line',
    'curve_to_compas_polyline',

    'xform_to_rhino',

    'RhinoGeometry',
    'RhinoBox',
    'RhinoCircle',
    'RhinoCone',
    'RhinoCurve',
    'RhinoCylinder',
    'RhinoEllipse',
    'RhinoLine',
    'RhinoMesh',
    'RhinoPlane',
    'RhinoPoint',
    'RhinoPolyline',
    'RhinoSphere',
    'RhinoSurface',
    'RhinoVector',
]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from numpy import asarray
from numpy import zeros
from numpy import cross
from numpy import bincount

from scipy.sparse import coo_matrix
from scipy.sparse import spdiags

from compas.geometry import dot_vectors
from compas.geometry import length_vector
from compas.geometry import cross_vectors

from compas.numerical import normrow
from compas.numerical import adjacency_matrix
from compas.numerical import degree_matrix
from compas.numerical import connectivity_matrix
from compas.numerical import face_matrix


__all__ = [
    'mesh_adjacency_matrix',
    'mesh_connectivity_matrix',
    'mesh_degree_matrix',
    'mesh_face_matrix',
    'mesh_laplacian_matrix',
    'trimesh_cotangent_laplacian_matrix',
    'trimesh_vertexarea_matrix'
]


def mesh_adjacency_matrix(mesh, rtype='array'):
    """Creates a vertex adjacency matrix from a Mesh datastructure.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    rtype : Literal['array', 'csc', 'csr', 'coo', 'list'], optional
        Format of the result.

    Returns
    -------
    array_like
        Constructed adjacency matrix.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_polyhedron(6)
    >>> A = mesh_adjacency_matrix(mesh)
    >>> type(A)
    <class 'numpy.ndarray'>

    >>> A = mesh_adjacency_matrix(mesh, rtype='csr')
    >>> type(A)
    <class 'scipy.sparse.csr.csr_matrix'>

    """
    key_index = mesh.key_index()
    adjacency = [[key_index[nbr] for nbr in mesh.vertex_neighbors(key)] for key in mesh.vertices()]
    return adjacency_matrix(adjacency, rtype=rtype)


def mesh_connectivity_matrix(mesh, rtype='array'):
    """Creates a connectivity matrix from a Mesh datastructure.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    rtype : Literal['array', 'csc', 'csr', 'coo', 'list'], optional
        Format of the result.

    Returns
    -------
    array_like
        Constructed connectivity matrix.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_polyhedron(6)
    >>> C = mesh_connectivity_matrix(mesh)
    >>> type(C)
    <class 'numpy.ndarray'>

    >>> C = mesh_connectivity_matrix(mesh, rtype='csr')
    >>> type(C)
    <class 'scipy.sparse.csr.csr_matrix'>

    >>> xyz = asarray(mesh.vertices_attributes('xyz'))
    >>> C = mesh_connectivity_matrix(mesh, rtype='csr')
    >>> uv = C.dot(xyz)

    """
    key_index = mesh.key_index()
    edges = [(key_index[u], key_index[v]) for u, v in mesh.edges()]
    return connectivity_matrix(edges, rtype=rtype)


def mesh_degree_matrix(mesh, rtype='array'):
    """Creates a vertex degree matrix from a Mesh datastructure.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    rtype : Literal['array', 'csc', 'csr', 'coo', 'list'], optional
        Format of the result.

    Returns
    -------
    array_like
        Constructed vertex degree matrix.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_polyhedron(6)
    >>> D = mesh_degree_matrix(mesh)
    >>> type(D)
    <class 'numpy.ndarray'>

    >>> D = mesh_degree_matrix(mesh, rtype='csr')
    >>> type(D)
    <class 'scipy.sparse.csr.csr_matrix'>

    >>> D = mesh_degree_matrix(mesh)
    >>> D.diagonal()
    array([3., 3., 3., 3., 3., 3., 3., 3.])

    """
    key_index = mesh.key_index()
    adjacency = [[key_index[nbr] for nbr in mesh.vertex_neighbors(key)] for key in mesh.vertices()]
    return degree_matrix(adjacency, rtype=rtype)


def mesh_face_matrix(mesh, rtype='array'):
    r"""Construct the face matrix from a Mesh datastructure.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    rtype : Literal['array', 'csc', 'csr', 'coo', 'list'], optional
        Format of the result.

    Returns
    -------
    array_like
        Constructed mesh face matrix.

    Notes
    -----
    The face matrix represents the relationship between faces and vertices.
    Each row of the matrix represents a face. Each column represents a vertex.
    The matrix is filled with zeros except where a relationship between a vertex
    and a face exist.

    .. math::

        F_{ij} =
        \begin{cases}
            1 & \text{if vertex j is part of face i} \\
            0 & \text{otherwise}
        \end{cases}

    The face matrix can for example be used to compute the centroids of all
    faces of a mesh.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_polyhedron(6)
    >>> F = mesh_face_matrix(mesh)
    >>> type(F)
    <class 'numpy.ndarray'>

    >>> F = mesh_face_matrix(mesh, rtype='csr')
    >>> type(F)
    <class 'scipy.sparse.csr.csr_matrix'>

    >>> from numpy import allclose
    >>> xyz = asarray(mesh.vertices_attributes('xyz'))
    >>> F = mesh_face_matrix(mesh, rtype='csr')
    >>> c1 = F.dot(xyz) / F.sum(axis=1)
    >>> c2 = [mesh.face_centroid(fkey) for fkey in mesh.faces()]
    >>> allclose(c1, c2)
    True

    """
    key_index = {key: index for index, key in enumerate(mesh.vertices())}
    face_vertices = [[key_index[key] for key in mesh.face_vertices(fkey)] for fkey in mesh.faces()]
    return face_matrix(face_vertices, rtype=rtype)


def mesh_laplacian_matrix(mesh, rtype='csr'):
    r"""Construct a Laplacian matrix with uniform weights from a mesh data structure.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    rtype : Literal['array', 'csc', 'csr', 'coo', 'list'], optional
        Format of the result.

    Returns
    -------
    array_like
        The Laplacian matrix.

    Notes
    -----
    The :math:`n \times n` uniform Laplacian matrix :math:`\mathbf{L}` of a mesh
    with vertices :math:`\mathbf{V}` and edges :math:`\mathbf{E}` is defined as
    follows [1]_

    .. math::

        \mathbf{L}_{ij} =
        \begin{cases}
            -1               & i = j \\
            \frac{1}{deg(i)} & (i, j) \in \mathbf{E} \\
            0                & \text{otherwise}
        \end{cases}

    with :math:`deg(i)` the degree of vertex :math:`i`.

    Therefore, the uniform Laplacian of a vertex :math:`\mathbf{v}_{i}` points to
    the centroid of its neighboring vertices.

    References
    ----------
    .. [1] Nealen A., Igarashi T., Sorkine O. and Alexa M.
        `Laplacian Mesh Optimization <https://igl.ethz.ch/projects/Laplacian-mesh-processing/Laplacian-mesh-optimization/lmo.pdf>`_.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_polyhedron(6)
    >>> L = mesh_laplacian_matrix(mesh, rtype='array')
    >>> type(L)
    <class 'numpy.ndarray'>

    >>> L = mesh_face_matrix(mesh, rtype='csr')
    >>> type(L)
    <class 'scipy.sparse.csr.csr_matrix'>

    >>> xyz = asarray(mesh.vertices_attributes('xyz'))
    >>> L = mesh_laplacian_matrix(mesh)
    >>> d = L.dot(xyz)

    """
    data, rows, cols = [], [], []
    key_index = mesh.key_index()

    for key in mesh.vertices():
        i = key_index[key]
        nbrs = mesh.vertex_neighbors(key)
        w = len(nbrs)
        data.append(-1.0)
        rows.append(i)
        cols.append(i)
        d = 1.0 / w
        for nbr in nbrs:
            j = key_index[nbr]
            data.append(d)
            rows.append(i)
            cols.append(j)

    L = coo_matrix((data, (rows, cols)))

    if rtype == 'csr':
        return L.tocsr()
    if rtype == 'csc':
        return L.tocsc()
    if rtype == 'array':
        return L.toarray()
    if rtype == 'list':
        return L.toarray().tolist()
    return L


def trimesh_edge_cotangent(mesh, u, v):
    """Compute the cotangent of the angle opposite a halfedge of the triangle mesh.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    u : int
        The identifier of the first vertex of the halfedge.
    v : int
        The identifier of the second vertex of the halfedge.

    Returns
    -------
    float
        The edge cotangent.

    """
    fkey = mesh.halfedge[u][v]
    cotangent = 0.0
    if fkey is not None:
        w = mesh.face_vertex_ancestor(fkey, u)
        wu = mesh.edge_vector(w, u)
        wv = mesh.edge_vector(w, v)
        length = length_vector(cross_vectors(wu, wv))
        if length:
            cotangent = dot_vectors(wu, wv) / length
    return cotangent


def trimesh_edge_cotangents(mesh, u, v):
    """Compute the cotangents of the angles opposite both sides of an edge of the triangle mesh.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.
    u : int
        The identifier of the first vertex of the edge.
    v : int
        The identifier of the second vertex of the edge.

    Returns
    -------
    tuple[float, float]
        The two edge cotangents.

    """
    a = trimesh_edge_cotangent(mesh, u, v)
    b = trimesh_edge_cotangent(mesh, v, u)
    return a, b


def trimesh_cotangent_laplacian_matrix(mesh, rtype='csr'):
    r"""Construct the Laplacian of a triangular mesh with cotangent weights.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        Instance of mesh.

    Returns
    -------
    array_like
        The Laplacian matrix with cotangent weights.

    Notes
    -----
    The cotangent laplacian of a vertex :math:`\mathbf{v}_{i}` points from the vertex
    to the projection of the vertex into the 1-ring plane.
    The cotangent laplacian vectors of a mesh thus provide an approximation of the
    per-vertex normals.

    The :math:`n \times n` cotangent Laplacian matrix :math:`\mathbf{L}` of a mesh
    with vertices :math:`\mathbf{V}` and edges :math:`\mathbf{E}` is defined as
    follows [1]_

    .. math::

        \mathbf{L}_{ij} =
            \begin{cases}
                -1      & \text{if i = j} \\
                w_{ij}  & \text{if (i, j) \in \mathbf{E}} \\
                0       & \text{otherwise}
            \end{cases}

    with

    .. math::

         w_{ij} = \frac{\omega_{ij}}{\sum_{(i, k) \in \mathbf{E}_{i}} \omega_{ik}}

    References
    ----------
    .. [1] Nealen A., Igarashi T., Sorkine O. and Alexa M.
        `Laplacian Mesh Optimization <https://igl.ethz.ch/projects/Laplacian-mesh-processing/Laplacian-mesh-optimization/lmo.pdf>`_.

    """
    key_index = mesh.key_index()
    n = mesh.number_of_vertices()
    data = []
    rows = []
    cols = []

    for key in mesh.vertices():
        nbrs = mesh.vertex_neighbors(key)
        i = key_index[key]
        data.append(-1.0)
        rows.append(i)
        cols.append(i)

        W = 0
        for nbr in nbrs:
            a, b = trimesh_edge_cotangents(mesh, key, nbr)
            w = a + b
            W += w

        for nbr in nbrs:
            j = key_index[nbr]
            a, b = trimesh_edge_cotangents(mesh, key, nbr)
            w = a + b
            data.append(w / W)
            rows.append(i)
            cols.append(j)

    L = coo_matrix((data, (rows, cols)), shape=(n, n))

    if rtype == 'csr':
        return L.tocsr()
    if rtype == 'csc':
        return L.tocsc()
    if rtype == 'array':
        return L.toarray()
    if rtype == 'list':
        return L.toarray().tolist()
    return L


def trimesh_positive_cotangent_laplacian_matrix(mesh):
    raise NotImplementedError


def trimesh_vertexarea_matrix(mesh):
    """Compute the n x n diagonal matrix of per-vertex voronoi areas.

    Parameters
    ----------
    mesh : :class:`compas.datastructures.Mesh`
        The triangle mesh data structure.

    Returns
    -------
    sparse matrix
        The diagonal voronoi area matrix.

    Examples
    --------
    >>> from compas.datastructures import Mesh
    >>> mesh = Mesh.from_polygons([[[0, 0, 0], [1, 0, 0], [0, 1, 0]]])
    >>> A = trimesh_vertexarea_matrix(mesh)
    >>> A.diagonal().tolist()
    [0.1666, 0.1666, 0.1666]

    """
    key_index = mesh.key_index()
    xyz = asarray(mesh.vertices_attributes('xyz'), dtype=float)
    tris = asarray([[key_index[key] for key in mesh.face_vertices(fkey)] for fkey in mesh.faces()], dtype=int)
    e1 = xyz[tris[:, 1]] - xyz[tris[:, 0]]
    e2 = xyz[tris[:, 2]] - xyz[tris[:, 0]]
    n = cross(e1, e2)
    a = 0.5 * normrow(n).ravel()
    a3 = a / 3.0
    area = zeros(xyz.shape[0])
    for i in (0, 1, 2):
        b = bincount(tris[:, i], a3)
        area[:len(b)] += b
    return spdiags(area, 0, xyz.shape[0], xyz.shape[0])



from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from compas.files.gltf.gltf_mesh import GLTFMesh
from compas.files.gltf.gltf_node import GLTFNode
from compas.files.gltf.gltf_scene import GLTFScene
from compas.files.gltf.helpers import get_weighted_mesh_vertices
from compas.geometry import transform_points
from compas.geometry import multiply_matrices
from compas.utilities import download_file_from_remote


class GLTFContent(object):
    """
    Class for managing the content of a glTF file.

    Attributes
    ----------
    scenes : dict
        Dictionary containing (int, :class:`compas.files.GLTFScene`) pairs.
    default_scene_key : int or None
        Key of the scene to be displayed on loading the glTF.
    nodes : dict
        Dictionary containing (int, :class:`compas.files.GLTFNode`) pairs.
    meshes : dict
        Dictionary containing (int, :class:`compas.files.GLTFMesh`) pairs.
    cameras : dict
        Dictionary containing (int, :class:`compas.files.data_classes.CameraData`) pairs.
    animations : dict
        Dictionary containing (int, :class:`compas.files.data_classes.AnimationData`) pairs.
    skins : dict
        Dictionary containing (int, :class:`compas.files.data_classes.SkinData`) pairs.
    materials : dict
        Dictionary containing (int, :class:`compas.files.data_classes.MaterialData`) pairs.
    textures : dict
        Dictionary containing (int, :class:`compas.files.data_classes.TextureData`) pairs.
    samplers : dict
        Dictionary containing (int, :class:`compas.files.data_classes.SamplerData`) pairs.
    images : dict
        Dictionary containing (int, :class:`compas.files.data_classes.ImageData`) pairs.
    extras : object
    extensions : object

    """
    def __init__(self):
        self.scenes = {}
        self.default_scene_key = None
        self.nodes = {}
        self.meshes = {}
        self.cameras = {}
        self.animations = {}
        self.skins = {}
        self.materials = {}
        self.textures = {}
        self.samplers = {}
        self.images = {}
        self.extras = None
        self.extensions = None

    @property
    def default_or_first_scene(self):
        key = self.default_scene_key or 0
        return self.scenes[key]

    def check_if_forest(self):
        """Raises an exception if :attr:`compas.files.GLTFContent.nodes` is not a disjoint
        union of rooted trees.

        Returns
        -------

        """
        visited_nodes = set()

        def visit(key):
            node = self.nodes[key]
            if key in visited_nodes:
                raise Exception('Nodes do not form a rooted forest.')
            visited_nodes.add(key)
            for child_key in node.children:
                visit(child_key)

        for scene in self.scenes.values():
            for node_key in scene.children:
                visit(node_key)

    def remove_orphans(self):
        """Removes orphaned objects.

        Returns
        -------

        """
        node_visit_log = {key: False for key in self.nodes}
        mesh_visit_log = {key: False for key in self.meshes}
        camera_visit_log = {key: False for key in self.cameras}
        material_visit_log = {key: False for key in self.materials}
        texture_visit_log = {key: False for key in self.textures}
        sampler_visit_log = {key: False for key in self.samplers}
        image_visit_log = {key: False for key in self.images}

        def visit_node(key):
            node = self.nodes[key]
            node_visit_log[key] = True
            if node.mesh_key is not None:
                mesh_visit_log[node.mesh_key] = True
            if node.camera is not None:
                camera_visit_log[node.camera] = True
            for child_key in node.children:
                visit_node(child_key)

        # walk through scenes and update visit logs of nodes, meshes, and cameras.
        for scene in self.scenes.values():
            for node_key in scene.children:
                visit_node(node_key)

        # remove unvisited nodes
        self._remove_unvisited(node_visit_log, self.nodes)

        # remove unvisited meshes
        self._remove_unvisited(mesh_visit_log, self.meshes)

        # remove unvisited cameras
        self._remove_unvisited(camera_visit_log, self.cameras)

        # remove animations referencing no existing nodes
        for animation_key, animation in self.animations.items():
            visited_sampler_keys = []
            for channel in animation.channels:
                if not node_visit_log[channel.target.node]:
                    animation.channels.remove(channel)
                else:
                    visited_sampler_keys.append(channel.sampler)
            animation.samplers_dict = {
                key: animation.samplers_dict[key]
                for key in animation.samplers_dict
                if key in visited_sampler_keys
            }
            if not animation.samplers_dict:
                del self.animations[animation_key]

        # remove skins referencing no existing nodes
        for key, skin_data in self.skins.items():
            for joint_key in skin_data.joints:
                if not node_visit_log[joint_key]:
                    skin_data.joints.remove(joint_key)
            if not skin_data.joints:
                del self.skins[key]

        # walk through existing meshes and update materials visit log
        for mesh in self.meshes.values():
            for primitive in mesh.primitive_data_list:
                if primitive.material is not None:
                    material_visit_log[primitive.material] = True

        # remove unvisited materials
        self._remove_unvisited(material_visit_log, self.materials)

        # walk through existing materials and update textures visit log
        for material in self.materials.values():
            if material.normal_texture is not None:
                texture_visit_log[material.normal_texture.index] = True
            if material.occlusion_texture is not None:
                texture_visit_log[material.occlusion_texture.index] = True
            if material.emissive_texture is not None:
                texture_visit_log[material.emissive_texture.index] = True
            if material.pbr_metallic_roughness is not None:
                if material.pbr_metallic_roughness.base_color_texture is not None:
                    texture_visit_log[material.pbr_metallic_roughness.base_color_texture.index] = True
                if material.pbr_metallic_roughness.metallic_roughness_texture is not None:
                    texture_visit_log[material.pbr_metallic_roughness.metallic_roughness_texture.index] = True

        # remove unvisited textures
        self._remove_unvisited(texture_visit_log, self.textures)

        # walk through existing textures and update visit logs of samplers and images
        for texture in self.textures.values():
            if texture.sampler is not None:
                sampler_visit_log[texture.sampler] = True
            if texture.source is not None:
                image_visit_log[texture.source] = True

        # remove unvisited samplers
        self._remove_unvisited(sampler_visit_log, self.samplers)

        # remove unvisited images
        self._remove_unvisited(image_visit_log, self.images)

    def _remove_unvisited(self, log, dictionary):
        for key, visited in log.items():
            if not visited:
                del dictionary[key]

    def update_node_transforms_and_positions(self):
        """Walks through all nodes and updates their transforms and positions.  To be used when
        scene or nodes have been added or the nodes' matrices or TRS attributes have been set or updated.

        Returns
        -------

        """
        for scene in self.scenes.values():
            self.update_scene_transforms_and_positions(scene)

    def update_scene_transforms_and_positions(self, scene):
        """Walks through the scene tree and updates transforms and positions.  To be used when
        nodes have been added or the nodes' matrices or TRS attributes have been set or updated.

        Parameters
        ----------
        scene : :class:`compas.files.GLTFScene`

        Returns
        -------

        """
        origin = [0, 0, 0]
        for node_key in scene.children:
            node = self.nodes[node_key]
            node.transform = node.matrix or node.get_matrix_from_trs()
            node.position = transform_points([origin], node.transform)[0]
            queue = [node_key]
            while queue:
                cur_key = queue.pop(0)
                cur = self.nodes[cur_key]
                for child_key in cur.children:
                    child = self.nodes[child_key]
                    child.transform = multiply_matrices(
                        cur.transform,
                        child.matrix or child.get_matrix_from_trs()
                    )
                    child.position = transform_points([origin], child.transform)[0]
                    queue.append(child_key)

    def get_node_faces(self, node):
        """Returns the faces of the mesh at ``node``, if any.

        Parameters
        ----------
        node : :class:`compas.files.GLTFNode`

        Returns
        -------
        list
        """
        mesh_data = self.meshes.get(node.mesh_key)
        if mesh_data is None:
            return None
        return mesh_data.faces

    def get_node_vertices(self, node):
        """Returns the vertices of the mesh at ``node``, if any.

        Parameters
        ----------
        node : :class:`compas.files.GLTFNode`

        Returns
        -------
        list
        """
        mesh_data = self.meshes.get(node.mesh_key)
        if mesh_data is None:
            return None
        if node.weights is None:
            return mesh_data.vertices
        return get_weighted_mesh_vertices(mesh_data, node.weights)

    def add_scene(self, name=None, extras=None):
        """Adds a scene to the content.

        Parameters
        ----------
        name : str
        extras : object

        Returns
        -------
        :class:`compas.files.GLTFScene`
        """
        return GLTFScene(self, name=name, extras=extras)

    def add_node_to_scene(self, scene, node_name=None, node_extras=None):
        """Creates a :class:`compas.files.GLTFNode` and adds this node to the children of ``scene``.

        Parameters
        ----------
        scene : :class:`compas.files.GLTFScene`
        node_name : str
        node_extras : object

        Returns
        -------
        :class:`compas.files.GLTFNode`
        """
        if scene not in self.scenes.values():
            raise Exception('Cannot find scene.')
        node = GLTFNode(self, node_name, node_extras)
        scene.children.append(node.key)
        return node

    def add_child_to_node(self, parent_node, child_name=None, child_extras=None):
        """Creates a :class:`compas.files.GLTFNode` and adds this node to the children of ``parent_node``.

        Parameters
        ----------
        parent_node : :class:`compas.files.GLTFNode`
        child_name : str
        child_extras : object

        Returns
        -------
        :class:`compas.files.GLTFNode`
        """
        child_node = GLTFNode(self, child_name, child_extras)
        parent_node.children.append(child_node.key)
        return child_node

    def add_mesh(self, mesh):
        """Creates a :class:`compas.files.GLTFMesh` object from a compas mesh, and adds this
        to the content.

        Parameters
        ----------
        mesh : :class:`compas.datastructures.Mesh`

        Returns
        -------
        :class:`compas.files.GLTFMesh`
        """
        return GLTFMesh.from_mesh(self, mesh)

    def add_mesh_to_node(self, node, mesh):
        """Adds an existing mesh to ``node`` if ``mesh`` is a valid mesh key, or through ``add_mesh`` creates and adds a
        mesh to ``node``.

        Parameters
        ----------
        node : :class:`compas.files.GLTFNode`
        mesh : Union[:class:`compas.datastructures.Mesh`, int]

        Returns
        -------
        :class:`compas.files.GLTFMesh`
        """
        if isinstance(mesh, int):
            mesh_data = self.meshes[mesh]
        else:
            mesh_data = self.add_mesh(mesh)
        node.mesh_key = mesh_data.key
        return mesh_data

    def get_nodes_from_scene(self, scene):
        """Returns dictionary of nodes in the given scene, without a specified root.

        Parameters
        ----------
        scene : :class:`compas.files.GLTFScene`

        Returns
        -------
        dict
        """
        node_dict = {}

        def visit(key):
            node_dict[key] = self.nodes[key]
            for child in self.nodes[key].children:
                visit(child)

        for child_key in scene.children:
            visit(child_key)

        return node_dict

    def get_scene_positions_and_edges(self, scene):
        """Returns a tuple containing a dictionary of positions and a list of tuples representing edges.

        Parameters
        ----------
        scene : :class:`compas.files.GLTFScene`

        Returns
        -------
        tuple
        """
        positions_dict = {'root': [0, 0, 0]}
        edges_list = []

        def visit(node, key):
            for child_key in node.children:
                positions_dict[child_key] = self.nodes[child_key].position
                edges_list.append((key, child_key))
                visit(self.nodes[child_key], child_key)

        visit(scene, 'root')

        return positions_dict, edges_list


# ==============================================================================
# Main
# ==============================================================================

if __name__ == '__main__':

    import os
    import compas

    from compas.datastructures import Mesh
    from compas.files.gltf.gltf import GLTF

    source = 'https://raw.githubusercontent.com/ros-industrial/abb/kinetic-devel/abb_irb6600_support/meshes/irb6640/visual/link_1.stl'
    stl_filepath = os.path.join(compas.APPDATA, 'data', 'meshes', 'ros', 'link_1.stl')

    download_file_from_remote(source, stl_filepath, overwrite=False)

    gltf_filepath = os.path.join(compas.APPDATA, 'data', 'gltfs', 'double_link_1.gltf')

    mesh = Mesh.from_stl(stl_filepath)
    cnt = GLTFContent()
    scene = cnt.add_scene()
    node_1 = scene.add_child(node_name='Node1')
    mesh_data = node_1.add_mesh(mesh)
    node_2 = node_1.add_child(child_name='Node2')
    node_2.translation = [0, 0, 5]
    node_2.add_mesh(mesh_data.key)

    gltf = GLTF(gltf_filepath)
    gltf.content = cnt
    gltf.export(embed_data=True)
